#include <GoApi/Logging.h>
#include <kApi/Utils/kUtils.h>
#include <kApi/Utils/kDateTime.h>

#if defined(K_WINDOWS)
#include <windows.h>    // For DebugApi.h::IsDebuggerPresent() anq DebugApi.h::OutputDebugStringA()
                        // TBD: This is normally picked up automatically by kUtils.h -> kApiDef.h -> kApiCfg.h,
                        // but for some reason its not automatically being picked up.
#endif

#define GOAPILOG_USEC_PER_MSEC 1000L

namespace Go
{

GoApiCppFx(kStatus) Logging::Go_Debug_Log(int line, 
                                          const kChar* fileName, 
                                          kLogOption options, 
                                          const kChar* source, 
                                          const kChar* format, ...)
{
    kLogArgs args;
    // GOS-2563 - Send in epoch time in ms for FE display.
    args.dateTime = kDateTime_Now();
    args.upTime = (k64u)((args.dateTime - kDATE_TIME_EPOCH_UNIX) / GOAPILOG_USEC_PER_MSEC); 
    args.options = options;
    args.source = source;
    // Won't be exactly the same -- but at least pretty close for the VS Output.
    // Some json can be quite long.
    kChar noPosParamFormat[512];
    kChar logMsgBuffer[512];

    if (format != nullptr)
    {
        // For now, we process the kVarArgList in here, but if we
        // need to in the future, we may need to alter the signature and pass along
        // kVarArgList to kLog().
        kVarArgList argList;
        kVarArgList_Start(argList, format);
        {
            if (GoLogIsGocUser(options))
            {
                // TODO: Stripping of positional parameters could probably be moved inside
                //       of GcLogger if we didn't want to handle stripping of positional 
                //       parameters here.
                // 
                // For user system messages, we have to use a format that doesn't have positional 
                // parameters in its format string to be compatible with ensuing kStrPrintvf's.
                //
                // GOS-3678 - Use the final log message with the arguments already parsed in because the
                //            ticket revealed potential problems when passing arguments to the FE client.
                kCheck(StripPositionalParameters(format, noPosParamFormat, kCountOf(noPosParamFormat)));
                kCheck(kStrPrintvf(logMsgBuffer, kCountOf(logMsgBuffer), noPosParamFormat, argList));
            }
            else
            {
                // If non-user system message, assume that positional paramters are not being used.
                kCheck(kStrPrintvf(logMsgBuffer, kCountOf(logMsgBuffer), format, argList));
            }
        }
        kVarArgList_End(argList);
    }
    else
    {
        kCheck(kStrPrintf(logMsgBuffer, kCountOf(logMsgBuffer), ""));
    }

    //----------------------------------------------------------------------
    // Populate the message to be logged by kLog and callbacks.
    //----------------------------------------------------------------------
    args.message = logMsgBuffer;

    // GOS-4608: This is a bit of a workaround, but log to firesync per new 
    //           logging options and with source to trigger kApiLib_InvokeLogListeners(),
    //           (of which GcLogger is a listener for public/private log handling).
    //           This is to ensure that kApiLogStatic::lock is always acquired 
    //           BEFORE GcLoggerClass::lock, and so that logging works for both
    //           Go_Debug_Log() based macros, and for direct calls to kLogf().
    kCheck(kLog(args.options, args.source, args.message));

    //----------------------------------------------------------------------
    // Perform logging if VS debugger is attached.
    //----------------------------------------------------------------------
#if defined(K_WINDOWS) && defined(_DEBUG)
    kChar presBuffer[1024];
    kCheck(FormatLogBuffer(args.dateTime, args.options, args.source, args.message,
        presBuffer, kCountOf(presBuffer)));

    char dBuffer[1024];
    // This logs to the Visual Studio Output window using a format that is hot-clickable 
    // for ease in debugging.
    if (IsDebuggerPresent())
    {
        kCheck(kStrPrintf(dBuffer, kCountOf(dBuffer), "%s(%d) : %s\n", fileName, line, presBuffer));
        OutputDebugStringA(dBuffer);
    }
#endif

    return kOK;
}

// This is based off of xkFormatLogOption().  We may want FireSync to provide access to  this for us.
GoApiCppFx(kStatus) Logging::FormatLogOption(kLogOption options, kChar* buffer, kSize capacity)
{
    const kChar* value = "";

    if (options & kLOG_OPTION_PLATFORM)
    {
        if (options & kLOG_OPTION_WARNING)          value = "Wrn Plt";
        else if (options & kLOG_OPTION_ERROR)       value = "Err Plt";
        else                                        value = "Inf Plt";
    }
    else if (options > 0) // application
    {
        if (options & kLOG_OPTION_WARNING)          value = "Wrn App";
        else if (options & kLOG_OPTION_ERROR)       value = "Err App";
        else                                        value = "Inf App";
    }

    kCheck(kStrCopy(buffer, capacity, value)); 

    return kOK;
}

// This is based off of xkFormatLogMessage().  We may want FireSync to provide access to  this for us.
//
// From FSS-1301 Sprint Notes:
//
//    [2000-01-01 00:00:00 UTC+8:00] Err App:0x03100040.kHxCamera: This is the log message.
//
//Where
//  Inf     Information
//  Err     Error
//  Wrn     Warning
//  Plt     Platform-specific
//  App     Application-specific
//  0xXXXX  Upper 32-bits of application options
//  Source  Source
//
GoApiCppFx(kStatus) Logging::FormatLogBuffer(kDateTime dateTime, kLogOption options, const kChar* source, const kChar* message,
                                          kChar* buffer, kSize capacity)
{
    kText64 logOptions = { 0 };
    kText64 dateTimeText = { 0 };
    kBool isSenderEmpty = (source == kNULL || kStrEquals("", source));
    k32u applicationOptions = (options & 0xFFFFFFFF00000000) >> 32;    // only upper 32-bits are preserved
    
    kCheck(FormatLogOption(options, logOptions, kCountOf(logOptions)));
    kCheck(kDateTime_Format(dateTime, kDATE_TIME_FORMAT_SIMPLE, dateTimeText, kCountOf(dateTimeText)));

    if (!kStrEquals("", logOptions))
    {
        if (!isSenderEmpty)
        {
            if (applicationOptions != 0)
            {
                kStrPrintf(buffer, capacity, "[%s] %s.%s (0x%X): %s", dateTimeText, logOptions, source, applicationOptions, message);
            }
            else
            {
                kStrPrintf(buffer, capacity, "[%s] %s.%s: %s", dateTimeText, logOptions, source, message);
            }
        }
        else
        {
            if (applicationOptions != 0)
            {
                kStrPrintf(buffer, capacity, "[%s] %s (0x%X): %s", dateTimeText, logOptions, applicationOptions, message);
            }
            else
            {
                kStrPrintf(buffer, capacity, "[%s] %s: %s", dateTimeText, logOptions, message);
            }
        }
    }
    else
    {
        if (!isSenderEmpty)
        {
            kStrPrintf(buffer, capacity, "[%s] %s: %s", dateTimeText, source, message);
        }
        else
        {
            kStrPrintf(buffer, capacity, "[%s] %s", dateTimeText, message);
        }
    }

    return kOK;
}

// The kStrPrintvf() used to create the log filter list does not handle
// positional parameters in the text. This function strips out the
// positional parameters from the text, if any.
// Caller must ensure the data area for the stripped string is large
// enough (ie. at least the same size as the original string)
GoApiCppFx(kStatus) Logging::StripPositionalParameters(const kChar* orig, kChar* stripped, kSize capacity)
{
    char* dstWrite;
    const char* srcFrom;
    const char* srcTo;
    const char* srcTemp;
    kSize numBytes;

    dstWrite = stripped;
    srcFrom = orig;
    srcTo = strstr(orig, "%");

    while (srcTo)
    {
        // Found a format specifier. Copy from original
        // up to and including the '%' character.
        srcTo++;
        numBytes = (kSize) (srcTo - srcFrom);
        kAssert(capacity >= numBytes);
        kCheckState(capacity >= numBytes);
        kMemCopy(dstWrite, srcFrom, numBytes);
        capacity -= numBytes;

        // Point to next location to write into.
        dstWrite += numBytes;

        // Point to char following the '%'.
        srcFrom = srcTo;
        srcTemp = srcTo;

        // Skip over any digits. Note that digits might be the width specifiers,
        // so need to look for a '$' to know if the digits are the positional
        // parameters or not.
        if (kChar_IsDigit(*srcTemp))
        {
            for (; kChar_IsDigit(*srcTemp); srcTemp++);

            if (*srcTemp == '$')
            {
                // Digits are for positional parameter so can skip the digits
                // and the '$'.
                srcTemp++;
                srcFrom = srcTemp;
                srcTo = srcTemp;
            }
        }

        // Find next format specifier, if any.
        srcTo = strstr(srcTo, "%");
    }

    // No more format specifiers. Copy the rest of the original string.
    for (; *srcFrom != '\0'; srcFrom++)
    {
        kAssert(capacity > 0);
        kCheckState(capacity > 0);

        *dstWrite = *srcFrom;
        dstWrite++;
        capacity--;
    }

    kAssert(capacity > 0);
    kCheckState(capacity > 0);
    *dstWrite = '\0';

    return kOK;
}

GoApiCppFx(std::string) Logging::StripTrailing(std::string str, std::string stripSet)
{
  std::size_t found = str.find_last_not_of(stripSet);
  if (found != std::string::npos)
  {
    // str consists entirely of chars in stripSet.
    return str.erase(found+1);
    
  }
  else
  {
    // str consists entirely of chars in stripSet.
    str.clear();
    return str;
  }
}

} // end namespace
