/**
 * @file    GoSystem.cpp
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
#include <GoPxLSdk/Def.h>
#include <GoPxLSdk/GoSystem.h>
#include <GoPxLSdk/GoRequestError.h>

namespace GoPxLSdk
{

const std::string CONTENT_PATH_EMBEDDED_ITEM = "/_embedded/item";
const std::string CONTENT_PATH_EMBEDDED_LINKS = "/_links/self/href";

const std::string SYSTEM_RESOURCE_PATH_ROOT = "/system";
const std::string SYSTEM_RESOURCE_PATH_START_CMD = "/system/commands/start";
const std::string SYSTEM_RESOURCE_PATH_STOP_CMD = "/system/commands/stop";
const std::string SYSTEM_RESOURCE_RUNSTATE_KEY = "runState";

const std::string SCAN_RESOURCE_PATH_ENGINES = "/scan/engines";
const std::string SCAN_RESOURCE_SUBPATH_SCANNERS = "/scanners";
const std::string SCAN_RESOURCE_SUBPATH_SENSORS = "/sensors";
const std::string SCAN_RESOURCE_SERIALNUM_KEY = "serialNumber";

GoSystem::GoSystem(kIpAddress address, k32u port) :
    address(address),
    port(port != 0 ? port : GO_PXL_SDK_DEFAULT_CONTROL_PORT)
{
    rest = std::make_shared<GoRestClient>();
}

void GoSystem::SetAddress(kIpAddress ipAddress)
{
    address = ipAddress;
}

kIpAddress GoSystem::Address() const
{
    return address;
}

void GoSystem::SetControlPort(k32u port)
{
    this->port = port;
}

k32u GoSystem::ControlPort() const
{
    return port;
}

void GoSystem::Connect()
{
    rest->Connect(address, port);
}

void GoSystem::Disconnect()
{
    rest->Disconnect();
}

bool GoSystem::IsConnected()
{
    return rest->IsConnected();
}

void GoSystem::Start()
{
    // GOS-4972: Add CheckResponse() to properly wait for the command to complete & throw if unsuccessful.
    rest->Call(SYSTEM_RESOURCE_PATH_START_CMD).CheckResponse(START_TIMEOUT_MSEC);
}

void GoSystem::Stop()
{
    // GOS-4972: Add CheckResponse() to properly wait for the command to complete & throw if unsuccessful.
    rest->Call(SYSTEM_RESOURCE_PATH_STOP_CMD).CheckResponse(STOP_TIMEOUT_MSEC);
}

GoSystem::State GoSystem::RunningState()
{
    return (State) rest->Read(SYSTEM_RESOURCE_PATH_ROOT).GetResponse(RUNNING_STATE_TIMEOUT_MSEC).Payload().Get<int>(SYSTEM_RESOURCE_RUNSTATE_KEY);
}

GoRestClient& GoSystem::Client()
{
    if (!rest)
    {
        throw GoRequestError();
    }

    return *rest;
}

ResourcePath GoSystem::SensorPath(SerialNum serialNum)
{
    ResourcePath enginesPath = SCAN_RESOURCE_PATH_ENGINES;
    std::vector<ResourcePath> engines = GetChildPaths(enginesPath);
    for (ResourcePath enginePath : engines)
    {
        ResourcePath scannersPath = enginePath + SCAN_RESOURCE_SUBPATH_SCANNERS;
        std::vector<ResourcePath> scanners = GetChildPaths(scannersPath);
        for (ResourcePath scannerPath : scanners)
        {
            ResourcePath sensorsPath = scannerPath + SCAN_RESOURCE_SUBPATH_SENSORS;
            std::vector<ResourcePath> sensors = GetChildPaths(sensorsPath);
            for (ResourcePath sensorPath : sensors)
            {
                SerialNum candidateSerialNum = rest->Read(sensorPath).GetResponse(SENSOR_PATH_TIMEOUT_MSEC).Payload().Get<SerialNum>(SCAN_RESOURCE_SERIALNUM_KEY);
                if (candidateSerialNum == serialNum)
                {
                    return sensorPath;
                }
            }
        }
    }

    // No matching serial number found.
    return "";
}

std::vector<ResourcePath> GoSystem::GetChildPaths(const ResourcePath& path)
{
    std::vector<ResourcePath> subPaths;

    try
    {
        // If there are no child paths (/_embedded/item doesn't exist), subItems will be an empty object.
        // GoJson::At will throw if key does not exist, so catch an GoJsonError and return an empty vector.
        const GoJson subItems = rest->Read(path).GetResponse(CHILD_PATH_TIMEOUT_MSEC).Payload().At(CONTENT_PATH_EMBEDDED_ITEM);

        for (GoJsonIterator i = subItems.Begin(); i != subItems.End(); i++)
        {
            ResourcePath subPath = i.Value().Get<std::string>(CONTENT_PATH_EMBEDDED_LINKS);
            subPaths.push_back(subPath);
        }

        return subPaths;
    }
    catch (const GoJsonError&)
    {
        return subPaths;
    }
}

}
