/**
 * @file    GoSystem.h
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
#ifndef GO_PXL_SDK_SYSTEM_H
#define GO_PXL_SDK_SYSTEM_H

#include <GoPxlSdk/Def.h>
#include <GoPxLSdk/GoRestClient.h>

#include <kApi/kApiDef.h>
#include <kApi/Io/kNetwork.h>

namespace GoPxLSdk
{
/**
 * Represents a single Gocator system.
 *
 * A Gocator system may have multiple sensors/devices.
 */
class GoPxLSdkClass GoSystem
{
/*
 * Constant member value declarations;
 */
private:

    /**
     * Constants used by operations available from this class to specify how long
     * to wait for a response from the requested operation. If no response to
     * the operation is received, then the operation fails with a timeout error.
     *
     * Define individual timeouts for each operation, rather than just one generic timeout
     * for all operations, because operations are not guaranteed to require the same amount
     * of time to complete.
     *
     * The GoSystem APIs are convenience functions that wrap around the underlying REST APIs.
     * Therefore, as convenience functions, they should be made to work as simply as possible
     * without needing the SDK user to configure or specifying anything to make these convenience
     * operations work.
     */
    static constexpr k64u START_TIMEOUT_MSEC            = 15000;
    static constexpr k64u STOP_TIMEOUT_MSEC             = 15000;
    static constexpr k64u RUNNING_STATE_TIMEOUT_MSEC    = 15000;
    static constexpr k64u SENSOR_PATH_TIMEOUT_MSEC      = 15000;
    static constexpr k64u CHILD_PATH_TIMEOUT_MSEC       = 15000;

public:
    enum class State
    {
        Ready = 0,
        Running = 1,
        Conflict = 2
    };

    /**
     * Constructs GoSystem.
     *
     * @public                              @memberof GoSystem
     * @version                             Introduced in 0.2.1.53.
     */
    GoSystem() = default;

    ~GoSystem() = default;
    /**
     * Constructs a GoSystem object with the given IP address and control port number.
     *
     * @public                              @memberof GoSystem
     * @version                             Introduced in 0.2.1.53.
     * @param address                       The IP address of the Gocator device to access.
     * @param port                          The IP control port number of the Gocator device. If 0, the default port will be used.
     */
    explicit GoSystem(kIpAddress address, k32u port = 0);

    /**
     * Sets the ipAddress used in Connect().
     *
     * @public                              @memberof GoSystem
     * @version                             Introduced in 0.2.1.53.
     */
    void SetAddress(kIpAddress ipAddress);

    /**
     * Gets the ipAddress used in Connect().
     *
     * @public                              @memberof GoSystem
     * @version                             Introduced in 0.2.1.53.
     */
    kIpAddress Address() const;

    /**
     * Sets the port used in Connect().
     *
     * @public                              @memberof GoSystem
     * @version                             Introduced in 0.2.1.53.
     */
    void SetControlPort(k32u port);

    /**
     * Gets the port used in Connect().
     *
     * @public                              @memberof GoSystem
     * @version                             Introduced in 0.2.1.53.
     */
    k32u ControlPort() const;

    /**
     * Connects to a system. Waits until the connection is established.
     *
     * @public                              @memberof GoSystem
     * @version                             Introduced in 0.2.1.53.
     * @throws Go::Exception                If the connection attempt failed.
     */
    void Connect();

    /**
     * Disconnects from a system. Waits until the disconnection is complete.
     *
     * @public                              @memberof GoSystem
     * @version                             Introduced in 0.2.1.53.
     * @throws Go::Exception                If unable to disconnect.
     */
    void Disconnect();

    /**
     * Checks if the system is connected or not.
     *
     * @public                              @memberof GoSystem
     * @version                             Introduced in 0.2.1.53.
     * @return                              True if connected, false otherwise.
     */
    bool IsConnected();

    /**
     * Starts the device.
     * This function throws a GoPxLSdk::GoChannelError if the operation is not completed within the timeout period.
     * Default timeout period is 15 seconds.
     *
     * @public                              @memberof GoSystem
     * @version                             Introduced in 0.2.1.53.
     * @throws GoPxLSdk::GoChannelError     If the request times out.
     * @throws GoPxlSdk::GoRequestError     If the server responds with a failed status for the request.
     */
    void Start();

    /**
     * Stops the device.
     * This function throws a GoPxLSdk::GoChannelError if the operation is not completed within the timeout period.
     * Default timeout period is 15 seconds.
     *
     * @public                              @memberof GoSystem
     * @version                             Introduced in 0.2.1.53.
     * @throws GoPxLSdk::GoChannelError     If the request times out.
     * @throws GoPxlSdk::GoRequestError     If the server responds with a failed status for the request.
     */
    void Stop();

    /**
     * Gets the system's current state.
     * This function throws a GoPxLSdk::GoChannelError if the operation is not completed within the timeout period.
     * Default timeout period is 15 seconds.
     *
     * @public                              @memberof GoSystem
     * @version                             Introduced in 0.2.1.53.
     * @throws GoPxLSdk::GoChannelError     If the request times out.
     * @throws GoPxlSdk::GoRequestError     If the server responds with a failed status for the request.
     * @return                              Current system state.
     */
    State RunningState();

    /**
     * Returns a reference to the client. Most operations are done directly through this client.
     *
     * @public                              @memberof GoSystem
     * @version                             Introduced in 0.2.1.53.
     * @throws GoPxlSd::GoRequestError      If no GoRestClient exist.       
     * @return                              GoRestClient reference.
     */ 
    GoRestClient& Client();

    /**
     * Given a sensor's serial number, returns its resource path.
     * This function throws a GoPxLSdk::GoChannelError if the operation is not completed within the timeout period.
     * Default timeout period is 15 seconds.
     *
     * @public                              @memberof GoSystem
     * @version                             Introduced in 0.2.1.53.
     * @param serialNum                     The serial number to look for.
     * @throws GoPxlSdk::GoRequestError     If the server responds with a failed status for the request.
     * @throws GoPxlSdk::GoChannelError     If the request times out.
     * @return                              Resource path of the sensor. If sensor with the specified serial number is not found, returns an empty string.
     */
    ResourcePath SensorPath(SerialNum serialNum);

// Private member variables.
private:
    std::shared_ptr<GoRestClient> rest;

    kIpAddress address = { };
    k32u port = GO_PXL_SDK_DEFAULT_CONTROL_PORT;

// Private member functions.
private:
    /**
     * Given a resource path, returns a list of its child paths.
     * This function throws a GoPxLSdk::GoRequestError if the path is invalid.
     * This function throws a GoPxLSdk::GoChannelError if the operation is not completed within the timeout period.
     * Default timeout period is 15 seconds.
     *
     * @param path                          The path for which we want the children resource paths.
     * @throws GoPxLSdk::GoRequestError     If path is invalid.
     * @throws GoPxLSdk::GoChannelError     If the request times out.
     * @return                              List of child paths.
     */
    std::vector<ResourcePath> GetChildPaths(const ResourcePath& path);
};

}

#endif
