/**
 * @file    Version.h
 * @brief   Declares build and SDK API version numbers.
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
#ifndef GOPXL_SDK_VERSION_H
#define GOPXL_SDK_VERSION_H

#include "Def.h"

namespace GoPxLSdk
{
// This version file describes the SDK build version digits.
// The build version is a 32-bit number made up of the digits as follows, where each digit
// occupies 8-bits in the version number:
//   version = MajorMinorReleaseBuild
#define GO_PXL_SDK_VERSION_MAJOR 8
#define GO_PXL_SDK_VERSION_MINOR 0
#define GO_PXL_SDK_VERSION_RELEASE 10
#define GO_PXL_SDK_VERSION_BUILD 42

// The SDK API version is the version of the SDK interfaces used to create SDK applications.
// This changes only if the SDK APIs change significantly for SDK applications, such as
// a change that breaks existing SDK applications or a change that adds new functionality
// for SDK applications to use.
// Note this version should match the GOREST_API_CURRENT_VERSION.
#define GO_PXL_SDK_API_VERSION_MAJOR    3
#define GO_PXL_SDK_API_VERSION_MINOR    0
#define GO_PXL_SDK_API_VERSION_PATCH    0

/**
 * Returns the SDK build version as a 32-bit number in the format:
 *    MMmmRRBB
 * where:
 *  MM = major number
 *  mm = minor number
 *  RR = release number
 *  BB = build number
 *
 * The SDK build version is the build version that generated the SDK package.
 *
 * @return          Gocator SDK build version number.
 */
GoPxLSdkCppFx(kVersion) BuildVersionNumber();

/**
 * Returns the SDK build version as a string. Format is:
 *    MM.mm.RR.BB
 * where:
 *  MM = major number
 *  mm = minor number
 *  RR = release number
 *  BB = build number
 * 
 * The SDK build version is the build version that generated the SDK package.
 *
 * @return          Gocator SDK build version string.
 */
GoPxLSdkCppFx(const kChar*) BuildVersion();

/**
 * Returns the SDK API version as string. Format is:
 *   MM.mm.PN
 * where:
 *  MM = major number
 *  mm = minor number
 *  PN = patch number
 *
 * The SDK API version is the version of the SDK interfaces used to create SDK applications.
 *
 * @return          Gocator SDK API version string.
 */
GoPxLSdkCppFx(const kChar*) ApiVersion();

}
#endif
