/*
 * @file            BackupRestore.cpp
 * Objective        Connect to a Gocator sensor system, save a backup archive, and restore a backup archive.
 * 
 * GoPxLSdk Sample
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 *
 * Licensed under the MIT License.
 * Redistributions of files must retain the above copyright notice.
 */

#include <fstream>
#include <filesystem>

#include <kApi/kApiDef.h>
#include <GoApi/GoApiLib.h>
#include <GoPxLSdk/GoDiscoveryClient.h>
#include <GoPxLSdk/GoResponse.h>
#include <GoPxLSdk/GoSystem.h>
#include <GoPxLSdk/GoChannelError.h>
#include <GoPxLSdk/GoRequestError.h>

using std::string;
using namespace GoPxLSdk;

const string SYSTEM_BACKUP_PATH         = "/system/commands/archive";
const string SYSTEM_RESTORE_PATH        = "/system/commands/restore";
const string BACKUP_FILE_PATH           = "./sample_backup.gpbak";
constexpr int ERROR_STATUS              = -1;
constexpr int OK_STATUS                 = 0;
// Ensure timeout value is set properly for your use case.
constexpr int REST_COMMAND_TIMEOUT_MSEC = 10'000;     // 10 seconds
constexpr int RECEIVE_DATA_TIMEOUT_MSEC = 20'000;     // Receiving data might require a higher timeout value.
constexpr const kChar* SENSOR_IP        = "192.168.1.10";

int BackupRestore()
{
    GoSystem system;
    kStatus status;
    kAssembly goApiLib = kNULL;
    GoJson response;
    GoJson payload;

    // Constructs GoPxL API core framework.
    if ((status = GoApiLib_Construct(&goApiLib)) != kOK)
    {
        std::cout << "Error: " << status << " - Failed to construct GoApiLib." << std::endl;
        return ERROR_STATUS;
    }

    // Connect to a sensor system.
    kIpAddress systemIpAddress;
    kIpAddress_Parse(&systemIpAddress, SENSOR_IP);
    system = GoSystem(systemIpAddress, GO_PXL_SDK_DEFAULT_CONTROL_PORT);

    kChar ipAddress[16];
    kIpAddress_Format(system.Address(), ipAddress, sizeof(ipAddress));

    try
    {
        system.Connect();
        std::cout << "\nConnected to " << ipAddress << ":" << system.ControlPort() << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Connection failed. Check if sensor is powered on and connected "
                                 "also check if you are using the correct IP address and port number."
                  << std::endl;
        return ERROR_STATUS;
    }

    // Fetch data and download as a file.
    payload = GoJson("{\"contents\" : [\"global\", \"allWorkspaces\", \"allJobs\", \"replay\"]}");
    // To save as a support file, use the following and change the BACKUP_FILE_PATH extension to.gpsup:
    // payload = "{\"contents\" : [\"global\", \"allWorkspaces\", \"allJobs\", \"replay\", \"support\"]}";

    // Get data.
    response = system.Client().Call(SYSTEM_BACKUP_PATH, payload).GetResponse(RECEIVE_DATA_TIMEOUT_MSEC).Payload();
    ByteArray data = response.At("data").GetBinary();
    std::cout << "Data received." << std::endl;

    try
    {
        // Download.
        char *buffer = (char *)data.data();
        std::ofstream dataFile(BACKUP_FILE_PATH, std::ios::out | std::ios::binary);
        dataFile.write(buffer, data.size());
        std::cout << "File downloaded to " << std::filesystem::absolute(BACKUP_FILE_PATH) << std::endl;
        dataFile.close();
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << "Failed to download the file to " << std::filesystem::absolute(BACKUP_FILE_PATH) << std::endl;
        return ERROR_STATUS;
    }

    // Upload and restore.
    std::cout << "\nRestoring system..." << std::endl;
    std::ifstream dataFile(BACKUP_FILE_PATH, std::ios::binary);
    std::vector<k8u> dataByteArray(std::istreambuf_iterator<char>(dataFile), {});

    auto payloadUpdated = GoJson(payload);
    payloadUpdated.Set("data", dataByteArray);
    // payloadUpdated.Set("data", data); // to restore using the currently received data.
    try
    {
        system.Client().Call(SYSTEM_RESTORE_PATH, payloadUpdated).CheckResponse(REST_COMMAND_TIMEOUT_MSEC);
        std::cout << "System restored using " << std::filesystem::absolute(BACKUP_FILE_PATH) << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Failed to restore the system using " << std::filesystem::absolute(BACKUP_FILE_PATH) << std::endl;
        return ERROR_STATUS;
    }

    system.Disconnect();

    return OK_STATUS;
}

int main(int argc, char **argv)
{
    int status;

    try
    {
        status = BackupRestore();
    }
    catch (const GoRequestError &e)
    {
        GoJson errors = e.GetResponse().Payload().At("errors");

        string errorMessage = "";

        if (!errors.Empty())
        {
            errorMessage.append(errors.At("/0/status").ToString());
            errorMessage.append(" ");
            errorMessage.append(errors.At("/0/description").ToString());
        }

        std::cerr << "GoRequestError: " << errorMessage << std::endl;
        std::cerr << "Error sending a REST command to " << e.GetResponse().Path() << std::endl;
        status = ERROR_STATUS;
    }
    catch (const GoChannelError &e)
    {
        std::cerr << "Error: " << e.what() << " - Check sensor status and make sure it is connected." << std::endl;
        status = ERROR_STATUS;
    }
    catch (const std::exception& e)
    {
        std::cerr << "Error: " << e.what() << std::endl;
        status = ERROR_STATUS;
    }

    std::cout << "\nPress ENTER to exit the program..." << std::endl;
    getchar();

    return status;
}
