/*
 * @file            ReceiveAsync.cpp
 * Objective        Receive data asynchronously.
 * 
 * GoPxLSdk Sample
 * Copyright (C) 2023 by LMI Technologies Inc.
 *
 * Licensed under the MIT License.
 * Redistributions of files must retain the above copyright notice.
 *
 * To receive data over Gocator Data Protocol (GDP), use GoGdpClient::ReceiveDataSync() or GoGdpClient::ReceiveDataAsync().
 *  GoGdpClient::ReceiveDataAsync() works as follows:
 *  1. GoGdpClient::connect() is called to establish connection with the Protocol.
 *  2. GoGdpClient::onDataReceive() starts a receiveThread & receives messages and deserializes them.
 *  3. Individual data messages are received until GoGdpClient::Close() is called.
 *  4. GoGdpClient::ReceiveDataAsync() receives individual data messages and re-assembles to GoDataSet object "dataSet".
 *  5. The new data set is added to the Go::MsgQueue<GoDataSet> dataQueue.
 *  6. In a second dataThread, the data set is removed from dataQueue.
 *  7. Callback function can be used to access dataSet.
 */

#include <kApi/kApiDef.h>
#include <GoApi/GoApiLib.h>
#include <GoPxLSdk/GoSystem.h>
#include <GoPxLSdk/GoChannelError.h>
#include <GoPxLSdk/GoRequestError.h>
#include <GoPxLSdk/GoGdpClient.h>
#include <GoPxLSdk/GoGdpMsg/GoGdpStamp.h>
#include <GoPxLSdk/GoGdpMsg/GoGdpMeasurement.h>

using std::string;
using namespace GoPxLSdk;

const string ENGINE_ID                  = "g2xxx";
const string SCANNER_PATH               = "/scan/engines/" + ENGINE_ID + "/scanners/scanner-0";
const string OUTPUT_STAMP_PATH          = "/scan/engines/" + ENGINE_ID + "/scanners/scanner-0/outputs/stamp";
const string GOCATOR_CONTROL_PATH       = "/controls/gocator";
const string GOCATOR_ADD_OUTPUT_PATH    = "/controls/gocator/outputs/commands/add";
constexpr int ERROR_STATUS              = -1;
constexpr int OK_STATUS                 = 0;
constexpr int REST_COMMAND_TIMEOUT_MSEC = 3'000; // 3 seconds.
constexpr const kChar* SENSOR_IP        = "192.168.1.10";

// Callback function for ReceiveDataAsync()
void onData(const GoDataSet &receivedDataSet)
{
    if (receivedDataSet.Count() < 1)
    {
        std::cout << "No data received." << std::endl;
        GoThrow(kERROR);
    }

    for (size_t i = 0; i < receivedDataSet.Count(); i++)
    {
        switch (receivedDataSet.GdpMsgAt(i).Type())
        {
        case GoPxLSdk::MessageType::STAMP:
        {
            const GoGdpStamp &stampMsg = (const GoGdpStamp &)receivedDataSet.GdpMsgAt(i);
            std::cout << "\nStamp" << std::endl;
            std::cout << "Frame index: " << stampMsg.FrameIndex() << std::endl;
            std::cout << "Timestamp: " << stampMsg.Timestamp() << std::endl;
            std::cout << "Encoder: " << stampMsg.Encoder() << std::endl;
            std::cout << "Encoder at Z: " << stampMsg.EncoderAtZ() << std::endl;
            std::cout << "Status: " << stampMsg.Status() << std::endl;
            break;
        }
        case GoPxLSdk::MessageType::MEASUREMENT:
        {
            const GoGdpMeasurement &MeasurementMsg = (const GoGdpMeasurement &)receivedDataSet.GdpMsgAt(i);
            std::cout << "\nMeasurement" << std::endl;
            std::cout << "Data Source: " << MeasurementMsg.DataSourceId() << std::endl;
            std::cout << "Array count: " << MeasurementMsg.ArrayedCount() << std::endl;
            std::cout << "Array index: " << MeasurementMsg.ArrayedIndex() << std::endl;
            std::cout << "Value: " << MeasurementMsg.Value() << std::endl;
            std::cout << "Decision: " << (int)MeasurementMsg.Decision() << std::endl;
            std::cout << "Output ID: " << MeasurementMsg.GdpId() << std::endl;
            break;
        }
        }
    }
}

int ReceiveAsync()
{
    GoSystem system;
    kStatus status;
    kAssembly goApiLib = kNULL;
    GoJson response;
    GoJson payload;

    // Constructs GoPxL API core framework.
    if ((status = GoApiLib_Construct(&goApiLib)) != kOK)
    {
        std::cout << "Error: " << status << " - Failed to construct GoApiLib." << std::endl;
        return ERROR_STATUS;
    }

    // Connect to a sensor system.
    kIpAddress systemIpAddress;
    kIpAddress_Parse(&systemIpAddress, SENSOR_IP);
    system = GoSystem(systemIpAddress, GO_PXL_SDK_DEFAULT_CONTROL_PORT);
    kChar ipAddress[16];
    kIpAddress_Format(system.Address(), ipAddress, sizeof(ipAddress));
    try
    {
        system.Connect();
        std::cout << "\nConnected to " << ipAddress << ":" << system.ControlPort() << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Connection failed. Check if sensor is powered on and connected "
                                 "also check if you are using the correct IP address and port number."
                  << std::endl;
        return ERROR_STATUS;
    }

    // Set trigger source and frame rate.
    const std::map<int, string> triggerSourceMap{{0, "Time"}, {1, "Encoder"}, {2, "External Input"}, {3, "Software"}};
    payload = GoJson("{\"parameters\" : {\"triggerSettings\" : {\"source\" : 0, \"frameRate\" : 0.3} }}");
    try
    {
        system.Client().Update(SCANNER_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT_MSEC);
        response = system.Client().Read(SCANNER_PATH).GetResponse().Payload();
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " Check the timeout value or the API path." << std::endl;
        return ERROR_STATUS;
    }

    // Run the sensor.
    if (system.RunningState() == GoSystem::State::Ready)
    {
        system.Start();
        std::cout << "\nSystem started." << std::endl;
    }

    // Enable Gocator Protocol.
    payload = GoJson("{\"enabled\": true}");
    system.Client().Update(GOCATOR_CONTROL_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT_MSEC);
    std::cout << "\nGocator Protocol enabled." << std::endl;

    // Adds stamp as a data output.
    response = system.Client().Read(OUTPUT_STAMP_PATH).GetResponse().Payload();
    string stampDataSourceId = response.At("dataSourceId").ToString();

    payload = GoJson("{\"source\":" + stampDataSourceId + ", \"outputId\":0, \"autoShift\":true}");
    system.Client().Call(GOCATOR_ADD_OUTPUT_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT_MSEC);
    std::cout << "\nStamp added to Gocator output." << std::endl;

    // Connect to the Gocator Data Protocol (GDP) server.
    auto gdpClient = std::make_unique<GoGdpClient>();
    try
    {
        gdpClient->Connect(system.Address(), GO_PXL_SDK_DEFAULT_GDP_SERVER_PORT);
        std::cout << "\nGocator Protocol connected." << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << "Error: " << e.what() << " - Failed to connect to the GDP server." << std::endl;
        return ERROR_STATUS;
    }

    // Establish an asynchronous data transfer and start receiving data.
    std::cout << "\nRunning callback function to receive data..." << std::endl;
    std::cout << "Press ENTER to close the connection and exit." << std::endl;
    std::function<void(const GoDataSet &)> func = onData;
    try
    {
        gdpClient->ReceiveDataAsync(func);
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Failed to receive data. Make sure data is available." << std::endl;
        return ERROR_STATUS;
    }

    getchar();
    gdpClient->Close();
    system.Stop();
    std::cout << "\nSystem stopped." << std::endl;
    system.Disconnect();
    return OK_STATUS;
}

int main(int argc, char **argv)
{
    int status;
    try
    {
        status = ReceiveAsync();
    }
    catch (const GoRequestError &e)
    {
        string errorMessage = "";

        const GoJson &payload = e.GetResponse().Payload();
        if (!payload.Empty() && payload.HasKey("errors"))
        {
            GoJson errors = payload.At("errors");
            errorMessage.append(errors.At("/0/status").ToString());
            errorMessage.append(" ");
            errorMessage.append(errors.At("/0/description").ToString());
        }
        std::cerr << "GoRequestError: " << errorMessage << std::endl;
        std::cerr << "Error sending a REST command to " << e.GetResponse().Path() << std::endl;
        status = ERROR_STATUS;
    }
    catch (const GoChannelError &e)
    {
        std::cerr << "Error: " << e.what() << " - Check sensor status and make sure it is connected." << std::endl;
        status = ERROR_STATUS;
    }
    catch (const std::exception& e)
    {
        std::cerr << "Error: " << e.what() << std::endl;
        status = ERROR_STATUS;
    }

    if (status == ERROR_STATUS)
    {
        std::cout << "\nPress ENTER to exit the program..." << std::endl;
        getchar();
    }

    return status;
}
