/*
 * @file            ReceiveMetrics.cpp
 * Objective        Connect to a Gocator sensor system and receive metrics.
 * 
 * GoPxLSdk Sample
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 *
 * Licensed under the MIT License.
 * Redistributions of files must retain the above copyright notice.
 */

#include <kApi/kApiDef.h>
#include <GoApi/GoApiLib.h>
#include <GoPxLSdk/GoSystem.h>
#include <GoPxLSdk/GoChannelError.h>
#include <GoPxLSdk/GoRequestError.h>

using std::string;
using namespace GoPxLSdk;

#define SENSOR_IP "192.168.1.10"

const string SYSTEM_METRICS_PATH = "/system/metrics";
const auto ERROR_STATUS = -1;
const auto OK_STATUS = 0;

// Callback function for SetStreamHandler().
static void onMetrics(const std::shared_ptr<GoStreamResponse> &notification)
{
    using std::cout, std::endl;
    GoJson content = notification->Payload();
    cout << "\nApplication Uptime: " << content.Get<int>("/appUpTime") << endl;
    cout << "CPU Cores Usage Average: " << content.Get<int>("/cpuCoresUsedAvg") << endl;
    cout << "Memory Capacity: " << content.Get<double>("/memCapacity") << endl;
    cout << "Memory Used: " << content.Get<double>("/memUsed") << endl;
    cout << "Type: " << notification->Type() << endl;
    cout << "Status: " << notification->Status() << endl;
    cout << "StreamId: " << notification->StreamIdentifier() << endl;
    cout << "StreamStatus: " << notification->StreamStatus() << endl;
    cout << "Path: " << notification->Path() << endl;
}

int ReceiveMetrics()
{
    GoSystem system;
    kStatus status;
    kAssembly goApiLib = kNULL;

    // Constructs GoPxL API core framework.
    if ((status = GoApiLib_Construct(&goApiLib)) != kOK)
    {
        std::cout << "Error: " << status << " - Failed to construct GoApiLib." << std::endl;
        return ERROR_STATUS;
    }

    // Connect to a sensor system.
    kIpAddress systemIpAddress;
    kIpAddress_Parse(&systemIpAddress, SENSOR_IP);
    system = GoSystem(systemIpAddress, GO_PXL_SDK_DEFAULT_CONTROL_PORT);

    kChar ipAddress[16];
    kIpAddress_Format(system.Address(), ipAddress, sizeof(ipAddress));

    try
    {
        system.Connect();
        std::cout << "\nConnected to " << ipAddress << ":" << system.ControlPort() << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Connection failed. Check if sensor is powered on and connected "
                                 "also check if you are using the correct IP address and port number."
                  << std::endl;
        return ERROR_STATUS;
    }

    /// Run the sensor.
    if (system.RunningState() == GoSystem::State::Ready)
    {
        system.Start();
        std::cout << "\nSystem started." << std::endl;
    }

    // Set the callback function and start receiving data.
    std::cout << "\nRunning callback function to receive metrics..." << std::endl;
    std::cout << "Press ENTER to close the connection and exit." << std::endl;
    try
    {
        system.Client().SetStreamHandler(onMetrics);
        system.Client().StartStream(SYSTEM_METRICS_PATH);
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Failed to receive data." << std::endl;
        return ERROR_STATUS;
    }

    getchar();
    system.Client().StopStream(SYSTEM_METRICS_PATH);
    system.Stop();
    std::cout << "\nSystem stopped." << std::endl;
    system.Disconnect();
    return OK_STATUS;
}

int main(int argc, char **argv)
{
    int status;
    try
    {
        status = ReceiveMetrics();
    }
    catch (const GoRequestError &e)
    {
        std::string errorMessage = "";

        const GoJson &payload = e.GetResponse().Payload();
        if (!payload.Empty() && payload.HasKey("errors"))
        {
            GoJson errors = payload.At("errors");
            errorMessage.append(errors.At("/0/status").ToString());
            errorMessage.append(" ");
            errorMessage.append(errors.At("/0/description").ToString());
        }

        std::cerr << "GoRequestError: " << errorMessage << std::endl;
        std::cerr << "Error sending a REST command to " << e.GetResponse().Path() << std::endl;
        status = ERROR_STATUS;
    }
    catch (const GoChannelError &e)
    {
        std::cerr << "Error: " << e.what() << " - Check sensor status and make sure it is connected." << std::endl;
        status = ERROR_STATUS;
    }
    catch (const std::exception& e)
    {
        std::cerr << "Error: " << e.what() << std::endl;
        status = ERROR_STATUS;
    }

    if (status == ERROR_STATUS)
    {
        std::cout << "\nPress ENTER to exit the program..." << std::endl;
        getchar();
    }
    return status;
}
