/*
 * @file            ReceiveProfile.cpp
 * Objective        Connect to the GoPxL Data Protocol and receive profile data.
 * 
 * GoPxLSdk Sample
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 *
 * Licensed under the MIT License.
 * Redistributions of files must retain the above copyright notice.
 *
 * GoGdpClient::ReceiveDataSync() or GoGdpClient::ReceiveDataAsync() can be used to receive data from the GoPxL Data Protocol (GDP) serve:
    1. Pass a timeout value (in microseconds) within which data will be received from the server.
    2. The GDP server sends out individual data messages. The last message is marked with an isLastMsg flag.
    3. ReceiveDataSync() receives individual data messages and reassembles to a data set (GoDataSet).
    4. GoGdpClient::DataSet() can be used to access the data set.
    5. GoGdpMsg::Type() indicates the type of individual messages in the dataset as described in GoPxLSdk::MessageType.
 */

#include <GoApi/GoApiLib.h>
#include <kApi/kApiDef.h>
#include <kApi/Data/kArray1.h>
#include <GoPxLSdk/GoSystem.h>
#include <GoPxLSdk/GoChannelError.h>
#include <GoPxLSdk/GoRequestError.h>
#include <GoPxLSdk/GoGdpClient.h>
#include <GoPxLSdk/GoGdpMsg/GoGdpStamp.h>
#include <GoPxLSdk/GoGdpMsg/GoGdpProfileUniform.h>
#include <GoPxLSdk/GoGdpMsg/GoGdpProfilePointCloud.h>

using std::string;
using namespace GoPxLSdk;

const string ENGINE_ID                      = "g2xxx";
const string OUTPUTS_PATH                   = "/scan/engines/" + ENGINE_ID + "/scanners/scanner-0/outputs";
const string OUTPUT_UNIFORM_PROFILE_PATH    = OUTPUTS_PATH + "/topUniformProfile";
const string GOCATOR_CONTROL_PATH           = "/controls/gocator";
const string GOCATOR_ADD_OUTPUT_PATH        = "/controls/gocator/outputs/commands/add";
constexpr int ERROR_STATUS                  = -1;
constexpr int OK_STATUS                     = 0;
constexpr int REST_COMMAND_TIMEOUT_MSEC     = 3'000; // 3 seconds.
constexpr int RECEIVE_DATA_TIMEOUT_MSEC     = 3'000;
constexpr const kChar* SENSOR_IP            = "192.168.1.10";

// Represents a point in a profile.
typedef struct ProfilePoint
{
    double x; // X coordinate - position along laser line (mm).
    double z; // Z coordinate - height at a given x position (mm).
    unsigned char intensity;
} ProfilePoint;

int ReceiveProfile()
{
    GoSystem system;
    kStatus status;
    kAssembly goApiLib = kNULL;
    GoJson response;
    GoJson payload;

    std::vector<ProfilePoint> profileBuffer;

    // Constructs GoPxL API core framework.
    if ((status = GoApiLib_Construct(&goApiLib)) != kOK)
    {
        std::cout << "Error: " << status << " - Failed to construct GoApiLib." << std::endl;
        return ERROR_STATUS;
    }

    // Connect to a sensor system.
    kIpAddress systemIpAddress;
    kIpAddress_Parse(&systemIpAddress, SENSOR_IP);
    system = GoSystem(systemIpAddress, GO_PXL_SDK_DEFAULT_CONTROL_PORT);

    kChar ipAddress[16];
    kIpAddress_Format(system.Address(), ipAddress, sizeof(ipAddress));

    try
    {
        system.Connect();
        std::cout << "\nConnected to " << ipAddress << ":" << system.ControlPort() << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Connection failed. Check if sensor is powered on and connected "
                                 "also check if you are using the correct IP address and port number."
                  << std::endl;
        return ERROR_STATUS;
    }

    // Run the sensor.
    if (system.RunningState() == GoSystem::State::Ready)
    {
        system.Start();
        std::cout << "\nSystem started." << std::endl;
    }

    // Enable Gocator Protocol.
    payload = GoJson("{\"enabled\": true}");
    try
    {
        system.Client().Update(GOCATOR_CONTROL_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT_MSEC);
        std::cout << "\nGocator Protocol enabled." << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " Check the timeout value or the API path." << std::endl;
        return ERROR_STATUS;
    }

    // Adds profile to Gocator Protocol output.
    payload = GoJson("{\"enabled\": true}");
    try
    {
        response = system.Client().Read(OUTPUT_UNIFORM_PROFILE_PATH).GetResponse().Payload();
        string profileDataSourceId = response.At("dataSourceId").ToString();
        payload = GoJson("{\"source\":" + profileDataSourceId + ", \"outputId\":0, \"autoShift\":true}");
        system.Client().Call(GOCATOR_ADD_OUTPUT_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT_MSEC);
        std::cout << "Added profile to the Gocator Protocol output." << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Failed to add profile to Gocator Protocol output. Check API path and ensure uniform spacing is enabled." << std::endl;
        return ERROR_STATUS;
    }

    // Connect to the Gocator Data Protocol (GDP) server.
    auto gdpClient = std::make_unique<GoGdpClient>();
    try
    {
        gdpClient->Connect(system.Address(), GO_PXL_SDK_DEFAULT_GDP_SERVER_PORT);
        std::cout << "\nGocator Protocol connected." << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << "Error: " << e.what() << " - Failed to connect to the GDP server." << std::endl;
        return ERROR_STATUS;
    }

    // Establish a synchronous data transfer.
    try
    {
        gdpClient->ReceiveDataSync(RECEIVE_DATA_TIMEOUT_MSEC);
        std::cout << "\nSynchronous data transfer established." << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Failed to receive data via synchronous connection." << std::endl;
        return ERROR_STATUS;
    }

    // Receive data.
    const GoDataSet &receivedDataSet = gdpClient->DataSet();
    if (receivedDataSet.Count() < 1)
    {
        std::cout << "No data is received." << std::endl;
        return ERROR_STATUS;
    }
    std::cout << "\nTotal number of messages: " << receivedDataSet.Count() << std::endl;

    // Loop through messages in the dataSet.
    for (size_t i = 0; i < receivedDataSet.Count(); i++)
    {
        std::cout << "\nMessage " << i + 1 << " of " << receivedDataSet.Count() << std::endl;
        std::cout << "Data source: " << receivedDataSet.GdpMsgAt(i).DataSourceId() << std::endl;

        switch (receivedDataSet.GdpMsgAt(i).Type())
        {

        case GoPxLSdk::MessageType::STAMP:
        {
            const GoGdpStamp &stampMsg = (const GoGdpStamp &)receivedDataSet.GdpMsgAt(i);
            std::cout << "\nStamp" << std::endl;
            std::cout << "Frame index: " << stampMsg.FrameIndex() << std::endl;
            std::cout << "Timestamp: " << stampMsg.Timestamp() << std::endl;
            std::cout << "Encoder: " << stampMsg.Encoder() << std::endl;
            std::cout << "Encoder at Z: " << stampMsg.EncoderAtZ() << std::endl;
            std::cout << "Status: " << stampMsg.Status() << std::endl;
            break;
        }

        case GoPxLSdk::MessageType::UNIFORM_PROFILE:
        {

            const GoGdpProfileUniform &uniformProfileMsg = (const GoGdpProfileUniform &)receivedDataSet.GdpMsgAt(i);
            std::cout << "\nUniform Profile" << std::endl;
            std::cout << "Date Source: " << uniformProfileMsg.DataSourceId() << std::endl;
            std::cout << "Array count: " << uniformProfileMsg.ArrayedCount() << std::endl;
            std::cout << "Array index: " << uniformProfileMsg.ArrayedIndex() << std::endl;

            // Transform to real world coordinates.
            int validPointCount = 0;
            int profilePointCount = uniformProfileMsg.Width();
            profileBuffer.resize(profilePointCount);

            /*  The uniform profile data is sent in an array of z values.
             *  The x real-world coordinates can be calculated from the index of the array,
             *  using the x offset and resolution.
             *  The z value must be transformed into real world coordinates before being used.
             */
            for (int k = 0; k < profilePointCount; k++)
            {
                k16s range;
                kArray1_Item(uniformProfileMsg.Ranges(), k, &range);

                profileBuffer[k].x = uniformProfileMsg.Offset().x + uniformProfileMsg.Resolution().x * k;

                if (range != k16S_NULL)
                {
                    // Calculates the Z position in mm from the given data.
                    // Z = Z offset + Z resolution * z
                    profileBuffer[k].z = uniformProfileMsg.Offset().z + uniformProfileMsg.Resolution().z * range;
                    validPointCount++;
                }
                else
                {
                    profileBuffer[k].z = k16S_NULL;
                }
            }
            std::cout << "Profile points count: " << profilePointCount << std::endl;
            std::cout << "Valid points count: " << validPointCount << std::endl;
            break;
        }

        case GoPxLSdk::MessageType::PROFILE_POINT_CLOUD:
        {

            const GoGdpProfilePointCloud &pointCloudProfileMsg = (const GoGdpProfilePointCloud &)receivedDataSet.GdpMsgAt(i);
            std::cout << "\nUniform Profile" << std::endl;
            std::cout << "Date Source: " << pointCloudProfileMsg.DataSourceId() << std::endl;
            std::cout << "Array count: " << pointCloudProfileMsg.ArrayedCount() << std::endl;
            std::cout << "Array index: " << pointCloudProfileMsg.ArrayedIndex() << std::endl;

            // Transform to real world coordinates.
            int validPointCount = 0;
            int profilePointCount = pointCloudProfileMsg.Width();
            profileBuffer.resize(profilePointCount);

            /*  The non-uniform profile data is sent in an array of x and z values.
             *  The x and z values must be transformed into real world coordinates,
             *  using the x and z offset and resolution values.
             */

            for (int k = 0; k < profilePointCount; k++)
            {
                kPoint16s point;
                kArray1_Item(pointCloudProfileMsg.Ranges(), k, &point);

                if (point.x != k16S_NULL)
                {
                    // Calculates the X and Z position in mm from the given data.
                    // X = X offset + X resolution * x
                    // Z = Z offset + Z resolution * z
                    profileBuffer[k].x = pointCloudProfileMsg.Offset().x + pointCloudProfileMsg.Resolution().x * point.x;
                    profileBuffer[k].z = pointCloudProfileMsg.Offset().z + pointCloudProfileMsg.Resolution().z * point.y;
                    validPointCount++;
                }
                else
                {
                    profileBuffer[k].x = k16S_NULL;
                    profileBuffer[k].z = k16S_NULL;
                }
            }

            std::cout << "Profile points count: " << profilePointCount << std::endl;
            std::cout << "Valid points count:" << validPointCount << std::endl;
            break;
        }

        default:
            std::cout << "\nNo Profile or Stamp received.\n"
                      << std::endl;
            break;
        }
    }

    getchar();
    gdpClient->Close();
    system.Stop();
    std::cout << "\nSystem stopped." << std::endl;
    system.Disconnect();
    return OK_STATUS;
}

int main(int argc, char **argv)
{
    int status;

    try
    {
        status = ReceiveProfile();
    }
    catch (const GoRequestError &e)
    {
        string errorMessage = "";

        GoJson payload = e.GetResponse().Payload();
        if (!payload.Empty())
        {
            GoJson errors = payload.At("errors");
            errorMessage.append(errors.At("/0/status").ToString());
            errorMessage.append(" ");
            errorMessage.append(errors.At("/0/description").ToString());
        }

        std::cerr << "GoRequestError: " << errorMessage << std::endl;
        std::cerr << "Error sending a REST command to " << e.GetResponse().Path() << std::endl;
        status = ERROR_STATUS;
    }
    catch (const GoChannelError &e)
    {
        std::cerr << "Error: " << e.what() << " - Check sensor status and make sure it is connected." << std::endl;
        status = ERROR_STATUS;
    }
    catch (const std::exception& e)
    {
        std::cerr << "Error: " << e.what() << std::endl;
        status = ERROR_STATUS;
    }

    return status;
}
