/*
 * @file            SaveJob.cpp
 * Objective        Connect to a Gocator sensor system, list saved jobs, save and load a new job.
 *
 * GoPxLSdk Sample
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 *
 * Licensed under the MIT License.
 * Redistributions of files must retain the above copyright notice.
 */

#include <fstream>
#include <filesystem>

#include <kApi/kApiDef.h>
#include <GoApi/GoApiLib.h>
#include <GoPxLSdk/GoResponse.h>
#include <GoPxLSdk/GoSystem.h>
#include <GoPxLSdk/GoChannelError.h>
#include <GoPxLSdk/GoRequestError.h>

using std::string;
using namespace GoPxLSdk;

const string VISIBLE_SENSORS_PATH       = "/scan/visibleSensors/";
const string JOBS_PATH                  = "/jobs";
const string JOB_FILES_PATH             = "/jobs/files";
const string SAVE_JOB_PATH              = "/jobs/commands/save";
const string LOAD_JOB_PATH              = "/jobs/commands/load";
const string RENAME_JOB_PATH            = "/jobs/commands/rename";
const string LOCAL_JOB_FILE_PATH        = "./sample_job.gpjob";
constexpr int ERROR_STATUS              = -1;
constexpr int OK_STATUS                 = 0;
constexpr int REST_COMMAND_TIMEOUT_MSEC = 3'000;      // 3 seconds.
constexpr int RECEIVE_DATA_TIMEOUT_MSEC = 20'000;     // Receiving data might require a higher timeout value.c
constexpr const kChar* SENSOR_IP        = "192.168.1.10";

int SaveJob()
{
    GoSystem system;
    kStatus status;
    kAssembly goApiLib = kNULL;
    GoJson response;
    GoJson payload;

    // Constructs GoPxL API core framework.
    if ((status = GoApiLib_Construct(&goApiLib)) != kOK)
    {
        std::cout << "Error: " << status << " - Failed to construct GoApiLib." << std::endl;
        return ERROR_STATUS;
    }

    // Connect to a sensor system.
    kIpAddress systemIpAddress;
    kIpAddress_Parse(&systemIpAddress, SENSOR_IP);
    system = GoSystem(systemIpAddress, GO_PXL_SDK_DEFAULT_CONTROL_PORT);
    kChar ipAddress[16];
    kIpAddress_Format(system.Address(), ipAddress, sizeof(ipAddress));
    try
    {
        system.Connect();
        std::cout << "\nConnected to " << ipAddress << ":" << system.ControlPort() << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Connection failed. Check if sensor is powered on and connected "
            "also check if you are using the correct IP address and port number."
            << std::endl;
        return ERROR_STATUS;
    }

    // Get the list of jobs.
    auto expandLevel = GoJson("{\"expandLevel\" : 1}");
    try
    {
        response = system.Client().Read(JOB_FILES_PATH, {}, expandLevel).GetResponse().Payload();
        if (response.HasKey("_embedded") && response.At("_embedded").HasKey("item")) {
            GoJson jobs = response.At("_embedded/item");
            std::cout << "\nList of saved jobs: " << std::endl;
            for (auto it = jobs.Begin(); it != jobs.End(); it.operator++(0))
            {
                std::cout << it.Value().At("jobName") << std::endl;
            }
        }
        else {
            std::cout << "\nCurrently GoPxL does not contain any saved job." << std::endl;
        }
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Failed to get jobs list."
            << std::endl;
    }
    
    // Save job.
    payload = GoJson("{\"name\" : \"SDK-demo\" }");
    system.Client().Call(SAVE_JOB_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT_MSEC);
    std::cout << "\nSaved the current job as " << payload.At("name") << std::endl;

    // Load job.
    response = system.Client().Read(JOBS_PATH, {}, expandLevel).GetResponse().Payload().At("/loadedJob");
    std::cout << "\nLoaded job: " << response.Get<string>() << std::endl;

    // Rename job.
    auto renameJobContent = GoJson("{ \"sourceName\" : \"SDK-demo\", \"destName\" : \"SDK-demo-job\" }");
    try
    {
        system.Client().Call(RENAME_JOB_PATH, renameJobContent).CheckResponse(REST_COMMAND_TIMEOUT_MSEC);
        std::cout << "\nJob renamed to " << renameJobContent.At("destName") << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Failed to rename the job. Make sure the job names are correct and the destination name does not already exist." << std::endl;
    }

    // Download job.
    auto READ_DATA_PATH = "/jobs/files/SDK-demo-job/data";
    response = system.Client().Read(READ_DATA_PATH, payload).GetResponse(RECEIVE_DATA_TIMEOUT_MSEC).Payload();
    ByteArray jobData = response.At("content").GetBinary();
    std::cout << "\nJob data received." << std::endl;
    try
    {
        char* buffer = (char*)jobData.data();
        std::ofstream recFile(LOCAL_JOB_FILE_PATH, std::ios::out | std::ios::binary);
        std::cout << "\nJob downloaded to " << std::filesystem::absolute(LOCAL_JOB_FILE_PATH) << std::endl;
        if (recFile.is_open()) {
            recFile.write(buffer, jobData.size());
            recFile.close();
        }
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << "Failed to download the job to " << std::filesystem::absolute(LOCAL_JOB_FILE_PATH) << std::endl;
        return ERROR_STATUS;
    }

    // Upload job.
    std::ifstream jobFile(LOCAL_JOB_FILE_PATH, std::ios::binary);
    try
    {
        if (jobFile.is_open()) {
            std::vector<k8u> jobData(std::istreambuf_iterator<char>(jobFile), {});
            GoJson localJobContent;
            localJobContent.Set("fromLive", false);
            localJobContent.Set("name", "SDK-local-job-demo");
            localJobContent.Set("content", jobData);

            system.Client().Create(JOB_FILES_PATH, localJobContent).CheckResponse(REST_COMMAND_TIMEOUT_MSEC);
            std::cout << "\nUpload job as " << localJobContent.At("name") << " from " << std::filesystem::absolute(LOCAL_JOB_FILE_PATH) << std::endl;
        }
        else {
            std::cerr << "\nFailed to upload file from " << std::filesystem::absolute(LOCAL_JOB_FILE_PATH) << std::endl;
        }
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Failed to upload the job. Make sure the job name does not already exist." << std::endl;
        return ERROR_STATUS;
    }
    
    system.Disconnect();
    return OK_STATUS;
}

int main(int argc, char** argv)
{
    int status;

    try
    {
        status = SaveJob();
    }
    catch (const GoRequestError& e)
    {
        string errorMessage = "";

        const GoJson& payload = e.GetResponse().Payload();
        if (!payload.Empty() && payload.HasKey("errors"))
        {
            GoJson errors = payload.At("errors");
            errorMessage.append(errors.At("/0/status").ToString());
            errorMessage.append(" ");
            errorMessage.append(errors.At("/0/description").ToString());
        }

        std::cerr << "GoRequestError: " << errorMessage << std::endl;
        std::cerr << "Error sending a REST command to " << e.GetResponse().Path() << std::endl;
        status = ERROR_STATUS;
    }
    catch (const GoChannelError& e)
    {
        std::cerr << "Error: " << e.what() << " - Check sensor status and make sure it is connected." << std::endl;
        status = ERROR_STATUS;
    }
    catch (const std::exception& e)
    {
        std::cerr << "Error: " << e.what() << std::endl;
        status = ERROR_STATUS;
    }

    std::cout << "\nPress ENTER to exit the program..." << std::endl;
    getchar();

    return status;
}