/**
 * @file    GoDiscoveryClient.cs
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
using MessagePack;
using Newtonsoft.Json;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.NetworkInformation;
using System.Net.Sockets;
using System.Runtime.InteropServices;
using System.Text;

namespace GoPxLSdk
{
    public class GoDiscoveryClient
    {
        public const ulong GOPXL_DISCOVERY_MESSAGE_ANNOUNCE = 0x1001;
        public const ulong GOPXL_BROADCAST_SIGNATURE = 0x4C58504F47494D4C; // "LMIGOPXL"

        private const int GOPXL_RESERVED_PORT_DISCOVERY_PROTOCOL = 3320;
        private const ulong GOPXL_DISCOVERY_MESSAGE_DISCOVER = 0x0001;
        private const long MAX_MESSAGE_SIZE = 1536;
        private const int CLIENT_BUFFER_SIZE = 1536;
        private const int CLIENT_SOCKET_SIZE = -1;

        private readonly List<GoInstance> _instances = new List<GoInstance>();

        /// <summary>Discovers all GoPxL instances on the network.</summary>
        /// <param name="timeoutMilliseconds">Time limit to listen for GoPxL instances.</param>
        /// <exception cref="InsufficientMemoryException">Thrown if buffer length is larger than maximum message size.</exception>
        public void BlockingDiscover(long timeoutMilliseconds)
        {
            UdpClient receiver = null;
            GoInstance info = null;
            var timer = new GoTimer();
            _instances.Clear();

            try
            {
                // Fill the message header that will be broadcasted to the server.
                DiscoveryBroadcastHeader header = new DiscoveryBroadcastHeader(
                    length: (ulong)Marshal.SizeOf(typeof(DiscoveryBroadcastHeader)),
                    messageId: GOPXL_DISCOVERY_MESSAGE_DISCOVER,
                    signature: GOPXL_BROADCAST_SIGNATURE);

                receiver = ConstructReceiver();
                Broadcast(header);

                timer.Start(timeoutMilliseconds);
                do
                {
                    byte[] buffer = Receive(receiver, (int)timer.Remaining, out IPEndPoint endpoint);
                    if (buffer.Length > 0)
                    {
                        if (buffer.Length > MAX_MESSAGE_SIZE)
                        {
                            throw new System.InsufficientMemoryException($"{nameof(GoDiscoveryClient)} received broadcast packet that exceeded max message size");
                        }

                        try
                        {
                            info = ParseReply(buffer);
                        }
                        catch
                        {
                            // We want to continue to receive UDP messages even if ParseReply fails
                            // to potentially parse more GoPxL instances.
                            continue;
                        }

                        if (info != null)
                        {
                            Console.WriteLine($"Track info, {endpoint}");
                            _instances.Add(info);
                        }
                    }
                }
                while (!timer.IsExpired);

            }
            finally
            {
                timer.Stop();
                receiver?.Dispose();
            }
        }

        /// <summary>Returns a list of all GoPxL instances found on the network.</summary>
        /// <returns>List of GoPxL instances.</returns>
        public List<GoInstance> InstanceList()
        {
            return _instances;
        }

        /// <summary>Returns a single GoPxL instance that is found on the network and has the specified 
        /// application Id.</summary>
        /// <param name="appId">Application Id of GoPxL instance to search for.</param>
        /// <returns>GoPxL instance that matches application Id.</returns>
        public GoInstance Instance(string appId)
        {
            return _instances.FirstOrDefault(i => i.AppId == appId);
        }

        private static byte[] Receive(UdpClient client, int millisecondsTimeout, out IPEndPoint remoteEP)
        {
            remoteEP = null;

            if (millisecondsTimeout <= 0)
            {
                return new byte[0];
            }

            var task = client.ReceiveAsync();
            bool completed = task.Wait(millisecondsTimeout);
            if (completed) 
            //var asyncResult = client.BeginReceive(null, null);
            //asyncResult.AsyncWaitHandle.WaitOne(millisecondsTimeout);

            //if (asyncResult.IsCompleted)
            {
                return task.Result.Buffer;
                //return client.EndReceive(asyncResult, ref remoteEP);
            }

            return new byte[0];
        }

        /// <summary>Constructs UdpClient object for receiving data.</summary>
        /// <param name="receiver">UdpClient to be constructed.</param>
        /// <exception cref="Exception">Thrown if failed to construct receiver.</exception>
        private UdpClient ConstructReceiver()
        {
            UdpClient receiver = null;
            try
            {
                receiver = new UdpClient(AddressFamily.InterNetwork)
                {
                    EnableBroadcast = true,
                };

                receiver.Client.ExclusiveAddressUse = false;
                receiver.Client.SetSocketOption(SocketOptionLevel.Socket, SocketOptionName.ReuseAddress, true);
                receiver.Client.Bind(new IPEndPoint(IPAddress.Any, GOPXL_RESERVED_PORT_DISCOVERY_PROTOCOL));

                return receiver;
            }
            catch (Exception ex)
            {
                receiver?.Dispose();
                throw new Exception("Failed to construct receiver.", ex);
            }
        }

        /// <summary>Constructs UdpClient object for sending data.</summary>
        /// <param name="sender">UdpClient to be constructed.</param>
        /// <param name="address">Ip address of local interface.</param>
        /// <exception cref="Exception">Thrown if failed to construct sender.</exception>
        private UdpClient ConstructSender(IPAddress address)
        {
            UdpClient sender = null;

            try
            {
                sender = new UdpClient(AddressFamily.InterNetwork)
                {
                    EnableBroadcast = true,
                    ExclusiveAddressUse = false
                };

                sender.Client.SetSocketOption(SocketOptionLevel.Socket, SocketOptionName.ReuseAddress, true);
                //sender.Client.ReceiveBufferSize = CLIENT_SOCKET_SIZE;
                sender.Client.Bind(new IPEndPoint(address, GOPXL_RESERVED_PORT_DISCOVERY_PROTOCOL));

                return sender;
            }
            catch (Exception ex)
            {
                sender?.Dispose();
                throw new Exception("Failed to construct sender.", ex);
            }
        }

        /// <summary>Broadcast the client message header to the broadcast port.</summary>
        /// <param name="data">Message header to be sent to port.</param>
        /// <param name="length">Size of message header.</param>
        /// <exception cref="Exception">Thrown if failed to broadcast message.</exception>
        private void Broadcast(DiscoveryBroadcastHeader header)
        {
            UdpClient sender = null;
            var data = header.ToByteArray();
            var remoteEP = new IPEndPoint(IPAddress.Broadcast, GOPXL_RESERVED_PORT_DISCOVERY_PROTOCOL);

            try
            {
                NetworkInterface[] adapters = NetworkInterface.GetAllNetworkInterfaces();
                // Filter to active IPv4 addresses
                var addresses = adapters
                    .Where(a => a.OperationalStatus == OperationalStatus.Up)
                    .SelectMany(a => a.GetIPProperties().UnicastAddresses)
                    .Where(a => a.Address.AddressFamily == AddressFamily.InterNetwork);

                // Include loopback addresses if using virtual devices.
                foreach (var localAddress in addresses)
                {
                    // We want to continue loop even if ConstructSender fails to create a sender for that 
                    // interface instead of skipping all remaining interfaces.
                    try
                    {
                        sender = ConstructSender(localAddress.Address);
                    }
                    catch
                    {
                        continue;
                    }
                    sender?.Send(data, data.Length, remoteEP);
                    sender?.Dispose();
                    sender = null;
                }
            }
            catch (System.Exception ex)
            {
                throw new Exception("Failed to broadcast message.", ex);
            }
            finally
            {
                sender?.Dispose();
            }
        }

        /// <summary>Convert the message received from the server into a GoInstance object and 
        /// save this object into the list of GoPxL instances found on the network.</summary>
        /// <param name="bytes">Message received from sender.</param>
        public GoInstance ParseReply(byte[] bytes)
        {
            // Extract the header of the message that was received from server.
            var serverHeader = DiscoveryServerHeader.Deserialize(bytes);

            // Ignore any discover messages.
            if (serverHeader.MessageId == GOPXL_DISCOVERY_MESSAGE_DISCOVER)
            {
                return null;
            }

            // Throw if an unexpected message
            if (serverHeader.MessageId != GOPXL_DISCOVERY_MESSAGE_ANNOUNCE
                || serverHeader.Signature != GOPXL_BROADCAST_SIGNATURE)
            {
                return null;
                // TODO: Why does the C++ sdk think this should be an exception/error?
                //throw new Exception("Unexpected server header message");
            }

            // Extract message payload.
            string content = Encoding.UTF8.GetString(bytes, DiscoveryServerHeader.StructSize, bytes.Length - DiscoveryServerHeader.StructSize);

            var instanceInfo = JsonConvert.DeserializeObject<GoInstance>(content);
            return instanceInfo;
        }

        /// <summary>
        /// Message request that is sent to the discovery server.
        /// </summary>
        [StructLayout(LayoutKind.Sequential, Pack = 1)] // Ensure tight packing
        private struct DiscoveryBroadcastHeader
        {
            public readonly ulong Length;
            public readonly ulong MessageId;
            public readonly ulong Signature;

            public DiscoveryBroadcastHeader(ulong length, ulong messageId, ulong signature)
            {
                Length = length;
                MessageId = messageId;
                Signature = signature;
            }

            public byte[] ToByteArray()
            {
                int size = Marshal.SizeOf(this);
                byte[] arr = new byte[size];

                IntPtr ptr = IntPtr.Zero;
                try
                {
                    ptr = Marshal.AllocHGlobal(size);
                    Marshal.StructureToPtr(this, ptr, true);
                    Marshal.Copy(ptr, arr, 0, size);
                }
                finally
                {
                    Marshal.FreeHGlobal(ptr);
                }

                return arr;
            }
        }

        /// <summary>
        /// Message header of the response that the discovery server sends back 
        /// to discovery client.
        /// </summary>
        [StructLayout(LayoutKind.Sequential, Pack = 1)] // Ensure tight packing
        private struct DiscoveryServerHeader
        {
            public static int StructSize => Marshal.SizeOf(typeof(DiscoveryServerHeader));

            public ulong Length;
            public ulong MessageId;
            public ulong Signature;
            public ulong Status;

            internal static DiscoveryServerHeader Deserialize(byte[] buffer)
            {
                GCHandle handle = GCHandle.Alloc(buffer, GCHandleType.Pinned);
                DiscoveryServerHeader header;
                try
                {
                    header = (DiscoveryServerHeader)Marshal.PtrToStructure(handle.AddrOfPinnedObject(), typeof(DiscoveryServerHeader));
                }
                finally
                {
                    handle.Free();
                }
                return header;
            }
        }
    }
}
