/**
 * @file    GoGdpImage.cs
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
using System;

namespace GoPxLSdk.GoGdpMsg
{
    public enum GoCfa
    {
        /// <summary>
        /// No color filter.
        /// </summary>
        None = 0,

        /// <summary>
        /// Bayer filter: BG/GR.
        /// </summary>
        BayerBGGR = 1,

        /// <summary>
        /// Bayer filter: GB/RG.
        /// </summary>
        BayerGBRG = 2,

        /// <summary>
        /// Bayer filter: RG/GB.
        /// </summary>
        BayerRGGB = 3,

        /// <summary>
        /// Bayer filter: GR/BG.
        /// </summary>
        BayerGRBG = 4,
    }

    public enum GoPixelFormat
    {
        /// <summary>
        /// Unknown pixel format.
        /// </summary>
        Null = 0,

        /// <summary>
        /// 8-bit greyscale
        /// </summary>
        Greyscale8BPP = 1,

        /// <summary>
        /// 8-bit color filter array
        /// </summary>
        Cfa8BPP = 2,

        /// <summary>
        /// 8-bits-per-channel color with 4 channels (blue/green/red/unused)(kRgb)
        /// </summary>
        Bgrx8BPC = 3,

        /// <summary>
        /// 1-bit greyscale, 8 packed pixels per image element
        /// </summary>
        Greyscale1BPP = 4,

        /// <summary>
        /// 16-bit greyscale
        /// </summary>
        Greyscale16BPP = 5,
    }

    public class GoGdpImage : GoGdpMsg
    {
        /// <summary>Get image height.</summary>
        /// <returns>image height value.</returns>
        public uint Height { get; internal set; } = 0;

        /// <summary>Get image width.</summary>
        /// <returns>image width value.</returns>
        public uint Width { get; internal set; } = 0;

        /// <summary>Get image pixel size.</summary>
        /// <returns>image pixel size.</returns>
        public uint PixelSize { get; internal set; } = 0;

        /// <summary>Get image pixel format.</summary>
        /// <returns>image pixel format.</returns>
        public GoPixelFormat PixelFormat { get; internal set; } = 0;

        /// <summary>Get image color filter.
        /// Color filter array alignment: 
        /// 0 - None/Null
        /// 1 - Bayer BG/GR 
        /// 2 - Bayer GB/RG 
        /// 3 - Bayer RG/GB 
        /// 4 - Bayer GR/BG</summary>
        /// <returns>color filter value.</returns>
        public GoCfa ColorFilter { get; internal set; } = GoCfa.None;

        /// <summary>Get image exposure.</summary>
        /// <returns>image exposure value in nanoseconds.</returns>
        public float Exposure { get; internal set; }

        /// <summary>Get flag to indicate if image is flipped about X-axis.</summary>
        /// <returns>true if flipped, false otherwise.</returns>
        public bool FlippedX { get; internal set; } = false;

        /// <summary>Get flag to indicate if image is flipped about Y-axis.</summary>
        /// <returns>true if flipped, false otherwise.</returns>
        public bool FlippedY { get; internal set; } = false;

        /// <summary>Get flag to indicate if the image is column based or not.</summary>
        /// <returns>true if column based, false otherwise.</returns>
        public bool ColumnBased { get; internal set; } = false;

        /// <summary>Get the image.</summary>
        /// <returns>kImage object.</returns>
        public byte[,] Image { get; internal set; }

        /// <summary>Constructs GoGdpImage.</summary>
        public GoGdpImage()
            : base(MessageType.Image)
        { }

        /// <summary>Deserialize video/image message.</summary>
        /// <param name="serializer">The serializer to read.</param>
        /// <exception cref="ArgumentOutOfRangeException">Thrown if failed to deserialize video/image message.</exception>
        internal override void Deserialize(GoBinaryReader serializer)
        {
            base.Deserialize(serializer);

            serializer.BeginRead<ushort>(true);

            try
            {
                Height = serializer.ReadUInt32();
                Width = serializer.ReadUInt32();
                PixelSize = serializer.ReadUInt32();
                PixelFormat = (GoPixelFormat)serializer.ReadInt32();
                ColorFilter = (GoCfa)serializer.ReadInt32();
                ushort reserved16 = serializer.ReadUInt16();

                Exposure = serializer.ReadSingle();

                FlippedX = serializer.ReadBool();
                FlippedY = serializer.ReadBool();
                ColumnBased = serializer.ReadBool();
            }
            finally
            {
                serializer.EndRead();
            }

            if ((PixelFormat == GoPixelFormat.Greyscale8BPP) || (PixelFormat == GoPixelFormat.Cfa8BPP))
            {
                Image = new byte[Height, Width]; // 1 byte per pixel
                var data = serializer.ReadBytes(Image.Length);
                System.Buffer.BlockCopy(data, 0, Image, 0, data.Length);
            }
            else if (PixelFormat == GoPixelFormat.Bgrx8BPC)
            {
                Image = new byte[Height, Width * 4]; // kRgb, 32u per pixel
                var data = serializer.ReadBytes(Image.Length);
                System.Buffer.BlockCopy(data, 0, Image, 0, data.Length);
            }
        }
    }
}