/**
 * @file    GoGdpMsg.cs
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
using System.Collections.Generic;

namespace GoPxLSdk.GoGdpMsg
{
    public abstract class GoGdpMsg
    {
        protected internal class CommonAttributes
        {
            public byte SpaceType { get; set; }
            public bool HasTransform { get; set; }
            public GoGdpTransform Transform { get; set; } = new GoGdpTransform();
            public bool HasBoundingBox { get; set; }
            public GoGdpBoundingBox BoundingBox { get; set; } = new GoGdpBoundingBox();
            public uint ArrayedCount { get; set; }
            public uint ArrayedIndex { get; set; }
            public string DataSourceId { get; set; }
            public string StampSourceId { get; set; }
            public ulong DataSetId { get; set; }
            public bool IsLastMsg { get; set; }
            public ushort GdpId { get; set; }

            public override bool Equals(object obj)
            {
                return obj is CommonAttributes attributes &&
                       SpaceType == attributes.SpaceType &&
                       HasTransform == attributes.HasTransform &&
                       Transform == attributes.Transform &&
                       HasBoundingBox == attributes.HasBoundingBox &&
                       BoundingBox == attributes.BoundingBox &&
                       ArrayedCount == attributes.ArrayedCount &&
                       ArrayedIndex == attributes.ArrayedIndex &&
                       DataSourceId == attributes.DataSourceId &&
                       StampSourceId == attributes.StampSourceId &&
                       DataSetId == attributes.DataSetId &&
                       IsLastMsg == attributes.IsLastMsg &&
                       GdpId == attributes.GdpId;
            }

            public override int GetHashCode()
            {
                int hashCode = 219206630;
                hashCode = hashCode * -1521134295 + SpaceType.GetHashCode();
                hashCode = hashCode * -1521134295 + HasTransform.GetHashCode();
                hashCode = hashCode * -1521134295 + EqualityComparer<GoGdpTransform>.Default.GetHashCode(Transform);
                hashCode = hashCode * -1521134295 + HasBoundingBox.GetHashCode();
                hashCode = hashCode * -1521134295 + EqualityComparer<GoGdpBoundingBox>.Default.GetHashCode(BoundingBox);
                hashCode = hashCode * -1521134295 + ArrayedCount.GetHashCode();
                hashCode = hashCode * -1521134295 + ArrayedIndex.GetHashCode();
                hashCode = hashCode * -1521134295 + EqualityComparer<string>.Default.GetHashCode(DataSourceId);
                hashCode = hashCode * -1521134295 + EqualityComparer<string>.Default.GetHashCode(StampSourceId);
                hashCode = hashCode * -1521134295 + DataSetId.GetHashCode();
                hashCode = hashCode * -1521134295 + IsLastMsg.GetHashCode();
                hashCode = hashCode * -1521134295 + GdpId.GetHashCode();
                return hashCode;
            }
        }

        /// <summary>Get concrete message type.</summary>
        /// <returns>gdp message type.</returns>
        public MessageType Type => _msgType;

        /// <summary>Get the data source Id.
        /// DataSource Id is used to identify the source of the message and
        /// distinguish multiple arrayed messages.</summary>
        /// <returns>const string reference for the data source Id.</returns>
        public string DataSourceId => Attributes.DataSourceId;

        /// <summary>Get the stamp source Id. Unique Id string used to identify the set that this message
        /// belongs to. As of writing, this is the "ScannerId", however in the future
        /// this will change to either the dataSourceId for the stamp output associated
        /// with this data or perhaps a more opaque uniuqe ID. Regardless, clients should
        /// treat this as an opaque ID, used solely for collecting many messages into a
        /// collection that describes a set.</summary>
        /// <returns>const string reference for the stamp source Id.</returns>
        public string StampSourceId => Attributes.StampSourceId;

        /// <summary>Get space type value. can be one of 
        /// 0 : none(Transform doesn't apply) 
        /// 1 : data 
        /// 2 : image
        /// </summary>
        /// <returns>space type value.</returns>
        public byte SpaceType => Attributes.SpaceType; 

        /// <summary>Get indicator if a transform (returned by Transform() API) exists for ths message. 
        /// If not, then the transform returned by Transform() API is an identity transform 
        /// which contains an identity matrix for the transform.</summary>
        /// <returns>true: if non-identify tranform is present, otherwise transform is an identity transform.</returns>
        public bool HasTransform => Attributes.HasTransform; 

        /// <summary>Get pointer to the transformation matrix values.
        /// Optional: only sent if HasTransform() returns true.
        /// Applying this transform to the associated data results in the data in common frame of reference.
        /// This is typically true when "spaceType" is set as 1 : data.
        /// If "spaceType" is on of the other values, (0:none, 2 : image, 3 : custom),
        /// the interpretation of how the transform is applied can be different.This is to be determined.</summary>
        /// <returns>reference to the transformation matrix values.</returns>
        public GoGdpTransform Transform => Attributes.Transform;

        /// <summary>Get the count of messages in the array. Any message could be a member of an arrayed
        /// output. If this is the case, this indicates the count of messages in the array. 
        /// If the message is not in an array (ie it is a non-arrayed output) the value of 
        /// this field will be 0.</summary>
        /// <returns>the count of messages in the array.</returns>
        public uint ArrayedCount => Attributes.ArrayedCount;

        /// <summary>Get the index of this message in the array. 
        /// This field only applies if arrayCount is >= 1.(if it is in an array).</summary>
        /// <returns>the index of this message in the array.</returns>
        public uint ArrayedIndex => Attributes.ArrayedIndex;

        /// <summary>Get indicator if a bounding box exists or not.</summary>
        /// <returns>true if bounding box exists, false if not.</returns>
        public bool HasBoundingBox => Attributes.HasBoundingBox;

        /// <summary>Get bounding box. 
        /// The bounding box is meaningful only if HasBoundingBox() returns true. 
        /// Otherwise the bounding box values are undefined.</summary>
        /// <returns>reference to the bounding box.</returns>
        public GoGdpBoundingBox BoundingBox => Attributes.BoundingBox;

        /// <summary>Get the data set identifier.
        /// The data set identifier is used to uniquely identify which set the current message belongs to.
        /// All messages within a set will always be received together.
        /// No messages between different sets will be interleaved.</summary>
        /// <returns>the message set idetnifier this message belongs to.</returns>
        public ulong DataSetId => Attributes.DataSetId;

        /// <summary>Get whether this messages is the last one within its current set.</summary>
        /// <returns>true if this is the last message within current set, false otherwise.</returns>
        public bool IsLastMsg => Attributes.IsLastMsg;

        /// <summary>Get the gdpId. 
        /// The gdpId is used to identify which data output source it is produced from.</summary>
        /// <returns>the gdpId.</returns>
        public ushort GdpId => Attributes.GdpId;

        protected internal CommonAttributes Attributes { get; } = new CommonAttributes();
        private readonly MessageType _msgType;

        /// <summary>Constructs GoGdpMsg.</summary>
        /// <param name="type">The message type.</param>
        protected GoGdpMsg(MessageType type = MessageType.NullType)
        {
            _msgType = type;
            Attributes.SpaceType = 0;

            Attributes.HasTransform = false;

            Attributes.HasBoundingBox = false;

            Attributes.ArrayedCount = 0;
            Attributes.ArrayedIndex = 0;

            Attributes.DataSetId = 0;
            Attributes.IsLastMsg = false;
            Attributes.GdpId = Def.NullType<ushort>();
        }

        /// <summary>Deserialize common attributes.</summary>
        /// <param name="serializer">The serializer to read.</param>
        /// <exception cref="ArgumentOutOfRangeException">Thrown if failed to deserialize common attributes.</exception>
        internal virtual void Deserialize(GoBinaryReader serializer)
        {
            // CommonAttrs
            serializer.BeginRead<uint>(true);
            try
            {
                Attributes.SpaceType = serializer.ReadByte();
                
                Attributes.HasTransform = serializer.ReadBool();

                if (Attributes.HasTransform)
                {
                    Attributes.Transform = new GoGdpTransform();
                    Attributes.Transform.Deserialize(serializer);
                }

                Attributes.HasBoundingBox = serializer.ReadBool();
                if (Attributes.HasBoundingBox)
                {
                    Attributes.BoundingBox = new GoGdpBoundingBox();
                    Attributes.BoundingBox.Deserialize(serializer);
                }

                Attributes.ArrayedCount = serializer.ReadUInt32();
                Attributes.ArrayedIndex = serializer.ReadUInt32();

                ushort length = serializer.ReadUInt16();
                Attributes.DataSourceId = Utils.ReadText(serializer, length);

                length = serializer.ReadUInt16();
                Attributes.StampSourceId = Utils.ReadText(serializer, length);

                Attributes.DataSetId = serializer.ReadUInt64();
                Attributes.IsLastMsg = serializer.ReadBool();

                Attributes.GdpId = serializer.ReadUInt16();
            }
            finally
            {
                serializer.EndRead();
            }
        }
    }
}