/**
 * @file    GoGdpSpots.cs
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
using System.Collections.Generic;

namespace GoPxLSdk.GoGdpMsg
{
    public class GdpSpot
    {
        public ushort Slice { get; }
        public uint Centre { get; }

        public GdpSpot(ushort slice, uint centre)
        {
            Slice = slice;
            Centre = centre;
        }

        public override bool Equals(object obj)
        {
            return obj is GdpSpot spot &&
                   Slice == spot.Slice &&
                   Centre == spot.Centre;
        }

        public override int GetHashCode()
        {
            int hashCode = 364034431;
            hashCode = hashCode * -1521134295 + Slice.GetHashCode();
            hashCode = hashCode * -1521134295 + Centre.GetHashCode();
            return hashCode;
        }
    }

    public class GoGdpSpots : GoGdpMsg
    {
        /// <summary>Get the count of points.</summary>
        /// <returns>the count of points.</returns>
        public uint PointCount { get; internal set; } = 0;

        /// <summary>Get exposure (ns).</summary>
        /// <returns>the exposure value.</returns>
        public float Exposure { get; internal set; } = 0.0f;

        /// <summary>Get the flag indicating spots are column-based or row based.</summary>
        /// <returns>the flag indicating column-based (true) or row-based (false).</returns>
        public bool ColumnBased { get; internal set; } = false;

        /// <summary>Get the scale value to apply to the slice to obtain slice position.</summary>
        /// <returns>the slice scale.</returns>
        public float SliceScale { get; internal set; } = 1.0f;

        /// <summary>Get the offset value to apply to the slice to obtain slice position.</summary>
        /// <returns>the slice offset.</returns>
        public float SliceOffset { get; internal set; } = 0;

        /// <summary>Get the scale value to apply to the center to obtain center position.</summary>
        /// <returns>the center scale.</returns>
        public float CenterScale { get; internal set; } = 1.0f;

        /// <summary>Get the offset value to apply to the center to obtain center position.</summary>
        /// <returns>the center offset.</returns>
        public float CenterOffset { get; internal set; } = 0;

        /// <summary>Get the Max slice count of the spots.</summary>
        /// <returns>the max slice count.</returns>
        public uint MaxSliceCount { get; internal set; } = 0;

        /// <summary>Get the minimum spot center value.</summary>
        /// <returns>the center minimum value.</returns>
        public uint SpotCenterMin { get; internal set; } = 0;

        /// <summary>Get the maximum spot center value.</summary>
        /// <returns>the center maximum value.</returns>
        public uint SpotCenterMax { get; internal set; } = 0;

        /// <summary>Get points representing slice and centre.</summary>
        /// <returns>the spot array reference.</returns>
        public IReadOnlyList<GdpSpot> Spots => _spots;

        private readonly List<GdpSpot> _spots = new List<GdpSpot>();

        /// <summary>Constructs GoGdpSpots.</summary>
        internal GoGdpSpots()
            : base(MessageType.Spots)
        { }

        internal void AddSpot(GdpSpot spot) => _spots.Add(spot);

        /// <summary>Deserialize spots message.</summary>
        /// <param name="serializer">The serializer to read.</param>
        /// <exception cref="ArgumentOutOfRangeException">Thrown if failed to deserialize spots message.</exception>
        internal override void Deserialize(GoBinaryReader serializer)
        {
            bool needEndRead = false;

            base.Deserialize(serializer);

            serializer.BeginRead<ushort>(true);
            needEndRead = true;
            try
            {
                PointCount = serializer.ReadUInt32();
                Exposure = serializer.ReadSingle();
                ColumnBased = serializer.ReadBool();
                SliceScale = serializer.ReadSingle();
                SliceOffset = serializer.ReadSingle();
                CenterScale = serializer.ReadSingle();
                CenterOffset = serializer.ReadSingle();
                MaxSliceCount = serializer.ReadUInt32();
                SpotCenterMin = serializer.ReadUInt32();
                SpotCenterMax = serializer.ReadUInt32();
                serializer.EndRead();
                needEndRead = false;

                for (int i = 0; i < PointCount; i++)
                {
                    var slice = serializer.ReadUInt16();
                    var centre = serializer.ReadUInt32();
                    _spots.Add(new GdpSpot(slice, centre));
                }
            }
            finally
            {
                if(needEndRead)
                {
                    serializer.EndRead();
                }
            }
        }
    }
}