/**
 * @file    GoGdpSurfacePointCloud.cs
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
using System;

namespace GoPxLSdk.GoGdpMsg
{
    public class GoGdpSurfacePointCloud : GoGdpSurfaceBase // GoPxLSdkClass, public GoGdpSurfaceBase
    {
        /// <summary>Check if the data is adjacent/sorted.</summary>
        /// <returns>true if is adjacant.</returns>
        public bool IsAdjacent { get; internal set; }

        /// <summary>Get surface ranges. Tuple(x, y, z)</summary>
        /// <returns>surface ranges array.</returns>
        public GoPoint3d16s[,] Ranges { get; internal set; }

        /// <summary>Get surface intensities.</summary>
        /// <returns>surface intensities array.</returns>
        public byte[,] Intensities { get; internal set; }

        /// <summary>Constructs GoGdpSurfacePointCloud.</summary>
        public GoGdpSurfacePointCloud()
            : base(MessageType.SurfacePointCloud)
        { }

        /// <summary>Deserialize surface point cloud message.</summary>
        /// <param name="serializer">The serializer to read.</param>
        /// <exception cref="ArgumentOutOfRangeException">Thrown if failed to deserialize surface point cloud message.</exception>
        internal override void Deserialize(GoBinaryReader serializer)
        {
            base.Deserialize(serializer);

            try
            {
                // Finish reading the surface attributes here.
                // This surface attribute is only sent for surface point clouds.
                // It is not applicable to uniform surfaces.

                IsAdjacent = serializer.ReadBool();
            }
            finally
            {
                // This closes the surface attributes read section.
                serializer.EndRead();
            }

            Ranges = new GoPoint3d16s[Length, Width];
            for (int j = 0; j < Length; j++)
            {
                for (int i = 0; i < Width; i++)
                {
                    // Incoming data is stored as an array of 16s
                    var x = serializer.ReadInt16();
                    var y = serializer.ReadInt16();
                    var z = serializer.ReadInt16();
                    Ranges[j, i] = new GoPoint3d16s(x, y, z);
            }
            }

            long intensitySize = IntensityLength * IntensityWidth;
            if (!serializer.IsReadCompleted && intensitySize > 0)
            {
                var intensitiesBuffer = serializer.ReadBytes((int)intensitySize);
                Intensities = new byte[IntensityLength, IntensityWidth];
                System.Buffer.BlockCopy(intensitiesBuffer, 0, Intensities, 0, intensitiesBuffer.Length);
            }
        }
    }
}