/**
 * @file    GoRequest.cs
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
using MessagePack;
using MessagePack.Resolvers;
using Newtonsoft.Json;
using Newtonsoft.Json.Converters;
using Newtonsoft.Json.Linq;
using Newtonsoft.Json.Serialization;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace GoPxLSdk
{
    [MessagePackObject]
    public class GoRequest
    {
        /// <summary>The request arguments.</summary>
        [JsonProperty("args")]
        [Key("args")]
        public JObject Args { get; private set; }

        /// <summary>The request method.</summary>
        [IgnoreMember]
        [JsonProperty("method")]
        [JsonConverter(typeof(StringEnumConverter))]
        public GoRequestMethod Method { get; private set; } = GoRequestMethod.Read;

        [Key("method")]
        public string MsgPackMethod { get => Method.ToString().ToLower(); set => throw new NotImplementedException(); }

        /// <summary>
        /// The relative URI of the request. 
        /// For example: 
        /// /scanners/A/sensors/0 
        /// /tool/someTool-0/inputs
        /// </summary>
        [Key("path")]
        [JsonProperty("path")]
        public string Path { get; private set; }

        /// <summary>
        /// Gets the request content. 
        /// Content are typically only used for Update, or less commonly, Create.
        /// </summary>
        [Key("payload")]
        [JsonProperty("payload")]
        public JObject Payload { get; private set; }

        /// <summary>Gets the request in form of ByteArray.</summary>
        /// <exception cref="MessagePackSerializationException">Thrown when any error occurs during serialization.</exception>
        public byte[] ToMsgPack()
        {
            var resolver = CompositeResolver.Create(
                // enable extension packages first
                JObjectResolver.Instance,
                //ContractlessStandardResolver.Instance,

                // finally use standard (default) resolver
                StandardResolver.Instance
            );
            var options = MessagePackSerializerOptions.Standard.WithResolver(resolver);

            return MessagePackSerializer.Serialize(this, options);
        }

        /// <summary>Gets the request in form of Json Array. But does not work yet as GoPxl SDK fully implemented or tested with JSON serialization </summary>
        /// <exception cref="NotImplementedException">Always thrown if called.</exception>
        public byte[] ToJsonArray()
        {
            throw new NotImplementedException();

            // This code should work, but as the .NET GoPxl SDK has not been fully implemented or tested with JSON serialization, 
            // this function has been disabled with the above throw to ensure that it is fast fails, and is
            // properly tested if we enable JSON functionality.

            //var json = JsonConvert.SerializeObject(this);
            //return Encoding.UTF8.GetBytes(json);
        }

        /// <summary>Constructs GoRequest.</summary>
        /// <param name="method">GoRequestMethod</param>
        /// <param name="path">Request path</param>
        public GoRequest(string method, string path)
            : this((GoRequestMethod)Enum.Parse(typeof(GoRequestMethod), method), path)
        { }

        /// <summary>Constructs GoRequest.</summary>
        /// <param name="method">GoRequestMethod</param>
        /// <param name="uri">Request path</param>
        public GoRequest(GoRequestMethod method, string path)
            : this(method, path, null, null)
        { }

        /// <summary>Constructs GoRequest.</summary>
        /// <param name="method">GoRequestMethod</param>
        /// <param name="uri">Request path</param>
        /// <param name="content">Request payload</param>
        public GoRequest(GoRequestMethod method, string path, JObject payload)
            : this(method, path, payload, null)
        { }

        /// <summary>Constructs GoRequest.</summary>
        /// <param name="method">GoRequestMethod</param>
        /// <param name="uri">Request path</param>
        /// <param name="content">Request payload</param>
        /// <param name="args">Request arguments</param>
        public GoRequest(GoRequestMethod method, string path, JObject payload, JObject args)
        {
            Method = method;
            Path = path;
            Payload = payload ?? new JObject();
            Args = args ?? new JObject();
        }
    }
}