/**
 * @file    GoResponse.cs
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
using MessagePack;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using System;

namespace GoPxLSdk
{
    [MessagePackObject]
    public class GoResponse
    {
        /// <summary>Returns the type of the response.</summary>
        [Key("type")]
        public GoResponseType Type { get; }

        /// <summary>Returns JSON message's payload field.</summary>
        [Key("payload")]
        public JObject Payload { get; }

        /// <summary>Returns JSON message's path field.</summary>
        [Key("path")]
        public string Path { get; }

        /// <summary>Returns full JSON message.</summary>
        [IgnoreMember]
        [JsonIgnore]
        public JObject Raw { get; }

        /// <summary>Returns JSON message status field.</summary>
        [IgnoreMember]
        public GoStatus Status => (GoStatus)_status;

        [Key("status")]
        private readonly int _status;

        /// <summary>Constructs GoResponse.</summary>
        public GoResponse(GoResponseType type, JObject fullmsg)
        {
            Raw = fullmsg;
            Type = type;
            Payload = fullmsg["payload"] as JObject;
            Path = fullmsg["path"]?.Value<string>() ?? null;

            // No enum parser to convert between GoStatus <-> int (like for Type) because the maintenance effort
            // is not justified. Instead, simple casting will be used to convert the int value obtained from
            // the JSON message into GoStatus enum type.
            _status = fullmsg["status"]?.Value<int>() ?? 0;
        }

        public GoResponse(string type, JObject fullmsg)
            : this((GoResponseType)Enum.Parse(typeof(GoResponseType), type, true), fullmsg)
        { }
    }

    /// <summary>
    /// Represents a response to requests such as Read/Update.
    /// </summary>
    [MessagePackObject]
    public sealed class GoRequestResponse : GoResponse
    {
        /// <summary>Constructs GoRequestResponse.</summary>
        public GoRequestResponse(JObject fullmsg)
            : base(GoResponseType.Request, fullmsg)
        { }
    }

    /// /// <summary>
    /// Represents change notifications message.
    /// </summary>
    [MessagePackObject]
    public sealed class GoNotificationResponse : GoResponse
    {
        /// <summary>Type of change from the resource.</summary>
        /// <returns>Notification type.</returns>
        [Key("eventType")]
        public GoNotificationType NotificationType { get; } = GoNotificationType.Changed;

        /// <summary>Constructs GoNotificationResponse.</summary>
        /// <param name="fullmsg">The JSON message.</param>
        public GoNotificationResponse(JObject fullmsg)
            : base(GoResponseType.Notification, fullmsg)
        {
            string notificationEventType = fullmsg["eventType"].Value<string>();

            NotificationType = (GoNotificationType)Enum.Parse(typeof(GoNotificationType), notificationEventType, true);
        }
    }

    /// <summary>
    /// Represents streamed data messages.
    /// </summary>
    [MessagePackObject]
    public sealed class GoStreamResponse : GoResponse
    {
        /// <summary>Returns the stream identifier number.</summary>
        [Key("streamId")]
        public uint StreamIdentifier { get; } = 0;

        /// <summary>Returns the stream status string.</summary>
        [Key("streamStatus")]
        public string StreamStatus { get; } = "";

        /// <summary>Constructs GoStreamResponse.</summary>
        /// <param name="fullmsg">The JSON message.</param>
        public GoStreamResponse(JObject fullmsg)
            : base(GoResponseType.Stream, fullmsg)
        {
            StreamIdentifier = Raw["streamId"]?.Value<uint>() ?? 0;
            StreamStatus = Raw["streamStatus"]?.Value<string>() ?? "";
        }
    }
}