/**
 * @file    GoSystem.cs
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
using System;
using System.Collections.Generic;
using System.Net;
using System.Text;

namespace GoPxLSdk
{
    /// <summary>
    /// Represents a single Gocator system.
    /// A Gocator system may have multiple sensors/devices.
    /// </summary>
    public class GoSystem : IDisposable
    {
        public enum State
        {
            Ready = 0,
            Running = 1,
            Conflict = 2
        }

        private const string CONTENT_PATH_EMBEDDED_ITEM = "$._embedded.item";
        private const string CONTENT_PATH_EMBEDDED_LINKS = "$._links.self.href";

        private const string SYSTEM_RESOURCE_PATH_ROOT = "/system";
        private const string SYSTEM_RESOURCE_PATH_START_CMD = "/system/commands/start";
        private const string SYSTEM_RESOURCE_PATH_STOP_CMD = "/system/commands/stop";
        private const string SYSTEM_RESOURCE_RUNSTATE_KEY = "runState";

        private const string SCAN_RESOURCE_PATH_ENGINES = "/scan/engines";
        private const string SCAN_RESOURCE_SUBPATH_SCANNERS = "/scanners";
        private const string SCAN_RESOURCE_SUBPATH_SENSORS = "/sensors";
        private const string SCAN_RESOURCE_SERIALNUM_KEY = "serialNumber";

        /// <summary>Constants used by operations available from this class to specify how long
        /// to wait for a response from the requested operation. If no response to
        /// the operation is received, then the operation fails with a timeout error.
        ///
        /// Define individual timeouts for each operation, rather than just one generic timeout
        /// for all operations, because operations are not guaranteed to require the same amount
        /// of time to complete.
        ///
        /// The GoSystem APIs are convenience functions that wrap around the underlying REST APIs.
        /// Therefore, as convenience functions, they should be made to work as simply as possible
        /// without needing the SDK user to configure or specifying anything to make these convenience
        /// operations work.</summary>
        private const int START_TIMEOUT_MSEC            = 15000;
        private const int STOP_TIMEOUT_MSEC             = 15000;
        private const int RUNNING_STATE_TIMEOUT_MSEC    = 15000;
        private const int SENSOR_PATH_TIMEOUT_MSEC      = 15000;
        private const int CHILD_PATH_TIMEOUT_MSEC       = 15000;

        /// <summary>Gets or sets the IP address used in <see cref="Connect"/>.</summary>
        public IPAddress Address { get; set; }

        /// <summary>Gets the port used in <see cref="Connect"/>.</summary>
        public int ControlPort { get; set; } = Def.GO_PXL_SDK_DEFAULT_CONTROL_PORT;

        /// <summary>Checks if the system is connected or not.</summary>
        /// <returns>True if connected, false otherwise.</returns>
        public bool IsConnected => _rest.IsConnected;

        private readonly GoRestClient _rest;

        /// <summary>
        /// Constructs a GoSystem object with the given IP address and control port number.
        /// </summary>
        /// <param name="address">The IP address of the Gocator device to access.</param>
        /// <param name="port">The IP control port number of the Gocator device. If 0, the default port will be used.</param>
        public GoSystem(IPAddress address, int port = 0)
        {
            _rest = new GoRestClient();
            Address = address;
            ControlPort = port != 0 ? port : Def.GO_PXL_SDK_DEFAULT_CONTROL_PORT;
        }

        /// <summary>Connects to a system. Waits until the connection is established.</summary>
        /// <exception cref="TimeoutException">Thrown if unable to connect.</exception>
        public void Connect()
        {
            _rest.Connect(Address, (uint) ControlPort);
        }

        /// <summary>Disconnects from a system. Waits until the disconnection is complete.</summary>
        /// <exception cref="Exception">Thrown if unable to disconnect.</exception>
        public void Disconnect()
        {
            _rest.Disconnect();
        }

        /// <summary>Starts the device.
        /// This function throws a GoPxLSdk::GoChannelError if the operation is not completed within the timeout period.
        /// Default timeout period is 15 seconds. </summary>
        /// <exception cref="GoRequestException">Thrown if the server responds with a failed status for the request.</exception>
        /// <exception cref="GoChannelException">Thrown if the request times out.</exception>
        public void Start()
        {
            // GOS-4972: Add CheckResponse() to properly wait for the command to complete & throw if unsuccessful.
            _rest.Call(SYSTEM_RESOURCE_PATH_START_CMD).CheckResponse(START_TIMEOUT_MSEC);
        }

        /// <summary>Stops the device.
        /// This function throws a GoPxLSdk::GoChannelError if the operation is not completed within the timeout period.
        /// Default timeout period is 15 seconds. </summary>
        /// <exception cref="GoRequestException">Thrown if the server responds with a failed status for the request.</exception>
        /// <exception cref="GoChannelException">Thrown if the request times out.</exception>
        public void Stop()
        {
            // GOS-4972: Add CheckResponse() to properly wait for the command to complete & throw if unsuccessful.
            _rest.Call(SYSTEM_RESOURCE_PATH_STOP_CMD).CheckResponse(STOP_TIMEOUT_MSEC);
        }

        /// <summary>Gets the system's current state.
        /// This function throws a GoPxLSdk::GoChannelError if the operation is not completed within the timeout period.
        /// Default timeout period is 15 seconds. </summary>
        /// <exception cref="GoRequestException">Thrown if the server responds with a failed status for the request.</exception>
        /// <exception cref="GoChannelException">Thrown if the request times out.</exception>
        /// <returns>Current system state.</returns>
        public State RunningState()
        {
            var transaction = _rest.Read(SYSTEM_RESOURCE_PATH_ROOT).GetResponse(RUNNING_STATE_TIMEOUT_MSEC);
            // TODO: how do we want to handle if the response (And thereby transaction payload) is in an error state? throw exception?
            var state = transaction.Payload[SYSTEM_RESOURCE_RUNSTATE_KEY].ToObject<State>();
            return state;
        }

        /// <summary>Returns a reference to the client. Most operations are done directly through this
        /// client.</summary>
        /// <exception cref="GoRequestException">Thrown if no GoRestClient exist.</exception>
        /// <returns>GoRestClient reference</returns>
        public GoRestClient Client()
        {
            if (_rest == null)
            {
                throw new GoRequestException();
            }

            return _rest;
        }
        /// <summary>Given a sensor's serial number, returns its resource path.
        /// This function throws a GoChannelError if the operation is not completed within the timeout period.
        /// Default timeout period is 15 seconds. </summary>
        /// <param name="serialNum">The serial number to look for.</param>
        /// <exception cref="GoRequestException">Thrown if the server responds with a failed status for the request.</exception>
        /// <exception cref="GoChannelException">Thrown if the request times out.</exception>
        /// <returns>resource path of the sensor. If sensor with the specified serial number is not found, returns an empty string.</returns>
        public string SensorPath(string serialNum)
        {
            string enginesPath = SCAN_RESOURCE_PATH_ENGINES;
            List<string> engines = GetChildPaths(enginesPath);
            foreach (string enginePath in engines)
            {
                string scannersPath = enginePath + SCAN_RESOURCE_SUBPATH_SCANNERS;
                List<string> scanners = GetChildPaths(scannersPath);
                foreach (string scannerPath in scanners)
                {
                    string sensorsPath = scannerPath + SCAN_RESOURCE_SUBPATH_SENSORS;
                    List<string> sensors = GetChildPaths(sensorsPath);
                    foreach (string sensorPath in sensors)
                    {
                        string candidateSerialNum = _rest.Read(sensorPath).GetResponse(SENSOR_PATH_TIMEOUT_MSEC).Payload[SCAN_RESOURCE_SERIALNUM_KEY].ToObject<string>();
                        if (candidateSerialNum == serialNum)
                        {
                            return sensorPath;
                        }
                    }
                }
            }

            // No matching serial number found.
            return "";
        }

        /// <summary>Given a resource path, returns a list of its child paths.
        /// This function throws a GoRequestException if the given path is invalid.
        /// This function throws a GoPxLSdk::GoChannelError if the operation is not completed within the timeout period.
        /// Default timeout period is 15 seconds. </summary>
        /// <param name="path">The path of which we want the children.</param>
        /// <exception cref="GoRequestException">Thrown if path is invalid.</exception>
        /// <exception cref="GoChannelException">Thrown if the request times out.</exception>
        /// <returns>list of child paths</returns>
        private List<string> GetChildPaths(string path)
        {
            List<string> subPaths = new List<string>();

            // If there are no child paths (/_embedded/item doesn't exist), subItems will be an empty object.
            var response = _rest.Read(path).GetResponse(CHILD_PATH_TIMEOUT_MSEC);
            var subItems = response.Payload.SelectToken(CONTENT_PATH_EMBEDDED_ITEM);
            if (subItems != null)
            {
                foreach (var item in subItems)
                {
                    string subPath = item.SelectToken(CONTENT_PATH_EMBEDDED_LINKS).ToObject<string>();
                    subPaths.Add(subPath);
                }
            }

            return subPaths;
        }

        public void Dispose()
        {
            Disconnect();
        }

        public override int GetHashCode()
        {
            int hashCode = -1281709372;
            hashCode = hashCode * -1521134295 + EqualityComparer<IPAddress>.Default.GetHashCode(Address);
            hashCode = hashCode * -1521134295 + ControlPort.GetHashCode();
            return hashCode;
        }
    }
}