/**
 * @file    DynamicEnumToLowerCaseStringResolver.cs
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
using MessagePack;
using MessagePack.Formatters;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;

namespace GoPxLSdk
{
    // Variant for forced lower case of Enum names. This can be deleted and replaced with [EnumMember] if upgrade to .NET Framework 7.3 or .NET 5/6
    internal sealed class DynamicEnumToLowerCaseStringResolver : IFormatterResolver
    {
        public static readonly DynamicEnumToLowerCaseStringResolver Instance;
        public static readonly MessagePackSerializerOptions Options;

        static DynamicEnumToLowerCaseStringResolver()
        {
            Instance = new DynamicEnumToLowerCaseStringResolver();
            Options = new MessagePackSerializerOptions(Instance);
        }

        private DynamicEnumToLowerCaseStringResolver() { }

        public IMessagePackFormatter<T> GetFormatter<T>() => FormatterCache<T>.Formatter;

        private static class FormatterCache<T>
        {
            public static readonly IMessagePackFormatter<T> Formatter;

            static FormatterCache()
            {
                TypeInfo ti = typeof(T).GetTypeInfo();

                bool isNullable = Nullable.GetUnderlyingType(ti) != null;
                if (isNullable)
                {
                    // build underlying type and use wrapped formatter.
                    ti = ti.GenericTypeArguments[0].GetTypeInfo();
                    if (!ti.IsEnum)
                    {
                        return;
                    }

                    var innerFormatter = Instance.GetFormatterDynamic(ti.AsType());
                    if (innerFormatter == null)
                    {
                        return;
                    }

                    Formatter = (IMessagePackFormatter<T>)Activator.CreateInstance(typeof(StaticNullableFormatter<>).MakeGenericType(ti.AsType()), new object[] { innerFormatter });
                    return;
                }
                else if (!ti.IsEnum)
                {
                    return;
                }

                Formatter = (IMessagePackFormatter<T>)(object)new EnumAsLowerCaseStringFormatter<T>();
            }
        }
    }
}
