/**
 * @file    GoBinaryReader.cs
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace GoPxLSdk
{
    internal class GoBinaryReader : BinaryReader
    {
        internal class ReadSection
        {
            public long Size { get; set; }
            public long Read { get; set; }
            public long HeaderSize { get; set; }
        }

        private readonly Stack<ReadSection> _sectionStack = new Stack<ReadSection>();

        public GoBinaryReader(Stream input)
            : base(input)
        { }

        public bool IsReadCompleted
        {
            get
            {
                var section = GetSection();
                if (section != null)
                    return section.Read == section.Size;
                else
                    return BaseStream.Position >= BaseStream.Length;
            }
        }

        public void BeginRead<T>(bool includeSize)
        {
            BeginRead(typeof(T), includeSize);
        }

        public void BeginRead(Type type, bool includeSize)
        {
            long size;

            if (type == typeof(byte))
            {
                size = ReadByte();
            }
            else if (type == typeof(ushort))
            {
                size = ReadUInt16();
            }
            else if (type == typeof(uint))
            {
                size = ReadUInt32();
            }
            else if (type == typeof(ulong))
            {
                size = (long)ReadUInt64();
            }
            else if (type == typeof(sbyte))
            {
                size = ReadSByte();
            }
            else if (type == typeof(short))
            {
                size = ReadInt16();
            }
            else if (type == typeof(int))
            {
                size = ReadInt32();
            }
            else if (type == typeof(long))
            {
                size = ReadInt64();
            }
            else if (type == typeof(float))
            {
                size = (long)ReadSingle();
            }
            else if (type == typeof(double))
            {
                size = (long)ReadDouble();
            }
            else
            {
                throw new NotImplementedException();
            }

            int read = 0;
            var headerSize = System.Runtime.InteropServices.Marshal.SizeOf(type);
            if (includeSize)
            {
                read = headerSize;
            }

            _sectionStack.Push(new ReadSection { HeaderSize = headerSize, Size = size, Read = read });
        }

        public void EndRead(bool exceptionIfDataRemaining = false)
        {
            if (_sectionStack.Count == 0)
                throw new Exception("EndRead was called, but no section is open for reading. Was BeginRead called?");

            var section = _sectionStack.Pop();
            long remaining = section.Size - section.Read;
            if (remaining > 0)
            {
                if (exceptionIfDataRemaining)
                    throw new Exception("EndRead was called, but there was still data remaining to be read in the section");

                BaseStream.Seek(remaining, SeekOrigin.Current);
            }

            if(_sectionStack.Count > 0)
            {
                var nextSection = _sectionStack.Peek();
                nextSection.Read += (section.Size - section.HeaderSize);
            }
        }

        internal ReadSection GetSection()
        {
            return _sectionStack.Count > 0 ? _sectionStack.Peek() : null;
        }

        public override bool ReadBoolean() => ReadBool();

        public bool ReadBool()
        {
            // Bools from FSS are saved as byte value 0/1
            byte byteValue = ReadByte();
            return byteValue > 0;
        }


        public GoPoint16s ReadGoPoint16s()
        {
            var section = GetSection();
            if (section != null)
            {
                int size = sizeof(short) * 2;
                if (section.Read + size > section.Size)
                    throw new ArgumentOutOfRangeException("Attempted to read data that exceeds the available open section of the reader.");

                section.Read += size;
            }

            // Use base, since we've already checked size for these doubles (don't want to double check/double increment read)
            short x = base.ReadInt16();
            short y = base.ReadInt16();
            return new GoPoint16s(x, y);
        }

        internal GoPoint3d16s ReadGoPoint3d16s()
        {
            var section = GetSection();
            if (section != null)
            {
                int size = sizeof(short) * 3;
                if (section.Read + size > section.Size)
                    throw new ArgumentOutOfRangeException("Attempted to read data that exceeds the available open section of the reader.");

                section.Read += size;
            }

            // Use base, since we've already checked size for these doubles (don't want to double check/double increment read)
            short x = base.ReadInt16();
            short y = base.ReadInt16();
            short z = base.ReadInt16();
            return new GoPoint3d16s(x, y, z);
        }

        internal GoPoint3d32f ReadGoPoint3d32f()
        {
            var section = GetSection();
            if (section != null)
            {
                int size = sizeof(float) * 3;
                if (section.Read + size > section.Size)
                    throw new ArgumentOutOfRangeException("Attempted to read data that exceeds the available open section of the reader.");

                section.Read += size;
            }

            // Use base, since we've already checked size for these doubles (don't want to double check/double increment read)
            float x = base.ReadSingle();
            float y = base.ReadSingle();
            float z = base.ReadSingle();
            return new GoPoint3d32f(x, y, z);
        }

        public GoPoint3d64f ReadGoPoint3d64f()
        {
            var section = GetSection();
            if (section != null)
            {
                int size = sizeof(double) * 3;
                if (section.Read + size > section.Size)
                    throw new ArgumentOutOfRangeException("Attempted to read data that exceeds the available open section of the reader.");

                section.Read += size;
            }

            // Use base, since we've already checked size for these doubles (don't want to double check/double increment read)
            double x = base.ReadDouble();
            double y = base.ReadDouble();
            double z = base.ReadDouble();
            return new GoPoint3d64f(x, y, z);
        }

        public override string ReadString()
        {
            var section = GetSection();
            long availableSpace = int.MaxValue;
            if (section != null)
            {
                availableSpace = section.Size - section.Read;
            }

            int size = sizeof(char);
            StringBuilder sb = new StringBuilder();

            availableSpace += size; // about to read a char, check space
            if (availableSpace < 0)
                throw new ArgumentOutOfRangeException("Attempted to read data that exceeds the available open section of the reader.");

            char next = base.ReadChar();
            section.Read += 1;
            while (next != 0)
            {
                sb.Append(next);

                availableSpace += size; // about to read a char, check space
                if (availableSpace < 0)
                    throw new ArgumentOutOfRangeException("Attempted to read data that exceeds the available open section of the reader.");

                next = base.ReadChar();
                section.Read += 1;
            }

            return sb.ToString();
        }

        public override byte ReadByte()
        {
            var section = GetSection();
            if (section != null)
            {
                int size = sizeof(byte);
                if (section.Read + size > section.Size)
                    throw new ArgumentOutOfRangeException("Attempted to read data that exceeds the available open section of the reader.");

                section.Read += size;
            }

            return base.ReadByte();
        }

        public override sbyte ReadSByte()
        {
            var section = GetSection();
            if (section != null)
            {
                int size = sizeof(sbyte);
                if (section.Read + size > section.Size)
                    throw new ArgumentOutOfRangeException("Attempted to read data that exceeds the available open section of the reader.");

                section.Read += size;
            }

            return base.ReadSByte();
        }

        public override char ReadChar()
        {
            var section = GetSection();
            if (section != null)
            {
                int size = sizeof(char);
                if (section.Read + size > section.Size)
                    throw new ArgumentOutOfRangeException("Attempted to read data that exceeds the available open section of the reader.");

                section.Read += size;
            }

            return base.ReadChar();
        }

        public override short ReadInt16()
        {
            var section = GetSection();
            if (section != null)
            {
                int size = sizeof(short);
                if (section.Read + size > section.Size)
                    throw new ArgumentOutOfRangeException("Attempted to read data that exceeds the available open section of the reader.");

                section.Read += size;
            }

            return base.ReadInt16();
        }

        public override int ReadInt32()
        {
            var section = GetSection();
            if (section != null)
            {
                int size = sizeof(int);
                if (section.Read + size > section.Size)
                    throw new ArgumentOutOfRangeException("Attempted to read data that exceeds the available open section of the reader.");

                section.Read += size;
            }

            return base.ReadInt32();
        }

        public override long ReadInt64()
        {
            var section = GetSection();
            if (section != null)
            {
                int size = sizeof(long);
                if (section.Read + size > section.Size)
                    throw new ArgumentOutOfRangeException("Attempted to read data that exceeds the available open section of the reader.");

                section.Read += size;
            }

            return base.ReadInt64();
        }

        public override ushort ReadUInt16()
        {
            var section = GetSection();
            if (section != null)
            {
                int size = sizeof(ushort);
                if (section.Read + size > section.Size)
                    throw new ArgumentOutOfRangeException("Attempted to read data that exceeds the available open section of the reader.");

                section.Read += size;
            }

            return base.ReadUInt16();
        }

        public override uint ReadUInt32()
        {
            var section = GetSection();
            if (section != null)
            {
                int size = sizeof(uint);
                if (section.Read + size > section.Size)
                    throw new ArgumentOutOfRangeException("Attempted to read data that exceeds the available open section of the reader.");

                section.Read += size;
            }

            return base.ReadUInt32();
        }

        public override ulong ReadUInt64()
        {
            var section = GetSection();
            if (section != null)
            {
                int size = sizeof(ulong);
                if (section.Read + size > section.Size)
                    throw new ArgumentOutOfRangeException("Attempted to read data that exceeds the available open section of the reader.");

                section.Read += size;
            }

            return base.ReadUInt64();
        }

        public override float ReadSingle()
        {
            var section = GetSection();
            if (section != null)
            {
                int size = sizeof(float);
                if (section.Read + size > section.Size)
                    throw new ArgumentOutOfRangeException("Attempted to read data that exceeds the available open section of the reader.");

                section.Read += size;
            }

            return base.ReadSingle();
        }

        public override double ReadDouble()
        {
            var section = GetSection();
            if (section != null)
            {
                int size = sizeof(double);
                if (section.Read + size > section.Size)
                    throw new ArgumentOutOfRangeException("Attempted to read data that exceeds the available open section of the reader.");

                section.Read += size;
            }

            return base.ReadDouble();
        }

        public override decimal ReadDecimal()
        {
            var section = GetSection();
            if (section != null)
            {
                int size = sizeof(decimal);
                if (section.Read + size > section.Size)
                    throw new ArgumentOutOfRangeException("Attempted to read data that exceeds the available open section of the reader.");

                section.Read += size;
            }

            return base.ReadDecimal();
        }

        public override int Read(char[] buffer, int index, int count)
        {
            var section = GetSection();
            if (section != null)
            {
                int size = count;
                if (section.Read + size > section.Size)
                    throw new ArgumentOutOfRangeException("Attempted to read data that exceeds the available open section of the reader.");

                section.Read += size;
            }

            return base.Read(buffer, index, count);
        }

        public override char[] ReadChars(int count)
        {
            var section = GetSection();
            if (section != null)
            {
                int size = count;
                if (section.Read + size > section.Size)
                    throw new ArgumentOutOfRangeException("Attempted to read data that exceeds the available open section of the reader.");

                section.Read += size;
            }

            return base.ReadChars(count);
        }

        public override int Read(byte[] buffer, int index, int count)
        {
            var section = GetSection();
            if (section != null)
            {
                int size = count;
                if (section.Read + size > section.Size)
                    throw new ArgumentOutOfRangeException("Attempted to read data that exceeds the available open section of the reader.");

                section.Read += size;
            }

            return base.Read(buffer, index, count);
        }

        public override byte[] ReadBytes(int count)
        {
            var section = GetSection();
            if (section != null)
            {
                int size = count;
                if (section.Read + size > section.Size)
                    throw new ArgumentOutOfRangeException("Attempted to read data that exceeds the available open section of the reader.");

                section.Read += size;
            }

            return base.ReadBytes(count);
        }
    }
}
