/**
 * @file    GoBinaryWriter.cs
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
using System;
using System.Collections.Generic;
using System.Diagnostics.Contracts;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace GoPxLSdk
{
    internal class GoBinaryWriter : BinaryWriter
    {
        private class BinaryWriteSection : GoBinaryWriter
        {
            public bool IncludeSize { get; }
            public Type Type { get; }
            public long Length => base.BaseStream.Length - (IncludeSize ? 0 : SizeOfType(this));
            public byte[] Data => ((MemoryStream)base.BaseStream).ToArray();

            public BinaryWriteSection(Type type, bool includeSize)
                : base(new MemoryStream())
            {
                Type = type;
                IncludeSize = includeSize;

                // reserve bytes at data start for writing length at endwrite
                BaseStream.Seek(SizeOfType(this), SeekOrigin.Begin);
            }

            private static int SizeOfType(BinaryWriteSection section)
            {
                if (section.Type == typeof(short))
                {
                    return sizeof(short);
                }
                else if (section.Type == typeof(ushort))
                {
                    return sizeof(ushort);
                }
                else if (section.Type == typeof(int))
                {
                    return sizeof(int);
                }
                else if (section.Type == typeof(uint))
                {
                    return sizeof(uint);
                }
                else if (section.Type == typeof(long))
                {
                    return sizeof(long);
                }
                else if (section.Type == typeof(ulong))
                {
                    return sizeof(ulong);
                }
                else
                {
                    throw new NotImplementedException();
                }
            }

            /// <summary>
            /// Write length to the base memory stream
            /// </summary>
            /// <remarks>
            /// We write the length in the subsection, as the root stream may be a stream with it's own ideas of how to write data. ie. for a network stream, 
            /// it can chop up consecutive write calls into separate network writes. The length should be written into the same body (data block).
            /// This is ensured by reserving a number of bytes in the subsection memorystream for the length value, and writing it to that reserved area before
            /// collapsing the subsection into it's parent.
            /// </remarks>
            public void WriteLength()
            {
                int currentPosition = (int)BaseStream.Position;
                Seek(0, SeekOrigin.Begin);

                try
                {
                    if (Type == typeof(short))
                    {
                        Write((short)Length);
                    }
                    else if (Type == typeof(ushort))
                    {
                        Write((ushort)Length);
                    }
                    else if (Type == typeof(int))
                    {
                        Write((int)Length);
                    }
                    else if (Type == typeof(uint))
                    {
                        Write((uint)Length);
                    }
                    else if (Type == typeof(long))
                    {
#pragma warning disable IDE0004
                        // Explicitly cast to long even though unnecessary. This guards against type changes in the future
                        Write((long)Length);
#pragma warning restore IDE0004
                    }
                    else if (Type == typeof(ulong))
                    {
                        Write((ulong)Length);
                    }
                    else
                    {
                        throw new NotImplementedException();
                    }
                }
                finally
                {
                    Seek(currentPosition, SeekOrigin.Begin);
                }
            }
        }

        private static readonly Encoding _encoding = Encoding.UTF8;
        private readonly Stack<BinaryWriteSection> _sectionStack = new Stack<BinaryWriteSection>();

        public GoBinaryWriter(Stream stream)
            : base(stream)
        { }

        public void BeginWrite<T>(bool includeSize) => BeginWrite(typeof(T), includeSize);

        public void BeginWrite(Type type, bool includeSize)
        {
            var section = new BinaryWriteSection(type, includeSize);
            _sectionStack.Push(section);
        }

        public void EndWrite()
        {
            if (_sectionStack.Count == 0)
            {
                throw new InvalidOperationException($"No section is in progress. Was {nameof(BeginWrite)} called?");
            }

            var section = _sectionStack.Pop();
            section.Flush();

            section.WriteLength();
            Write(section.Data);

            section.Close();
            section.Dispose();
        }

        private BinaryWriter GetSection() => _sectionStack.Count > 0 ? _sectionStack.Peek() : this;

        public new void Write(bool value)
        {
            // Bools from FSS are saved as byte value 0/1
            byte fssValue = value ? (byte)1 : (byte)0;
            GetSection().Write(fssValue);
        }

        public new void Write(byte value)
        {
            GetSection().Write(value);
        }

        public new void Write(sbyte value)
        {
            GetSection().Write(value);
        }

        public new void Write(byte[] buffer)
        {
            GetSection().Write(buffer);
        }

        public new void Write(byte[] buffer, int index, int count)
        {
            GetSection().Write(buffer, index, count);
        }

        public new void Write(char ch)
        {
            GetSection().Write(ch);
        }

        public new void Write(char[] chars)
        {
            GetSection().Write(chars);
        }

        public new void Write(char[] chars, int index, int count)
        {
            GetSection().Write(chars, index, count);
        }

        public new void Write(double value)
        {
            GetSection().Write(value);
        }

        public new void Write(decimal value)
        {
            GetSection().Write(value);
        }

        public new void Write(short value)
        {
            GetSection().Write(value);
        }

        public new void Write(ushort value)
        {
            GetSection().Write(value);
        }

        public new void Write(int value)
        {
            GetSection().Write(value);
        }

        public new void Write(uint value)
        {
            GetSection().Write(value);
        }

        public new void Write(long value)
        {
            GetSection().Write(value);
        }

        public new void Write(ulong value)
        {
            GetSection().Write(value);
        }

        public new void Write(float value)
        {
            GetSection().Write(value);
        }

        // C# normally does a length-prefixed string marking. FSS uses a null-terminated marker.
        // We override the default C# string write to use the FSS style for networking compatability.
        public new void Write(string value) => Write(value, true);

        public void Write(string value, bool nullTerminate)
        {
            if (value == null)
                throw new ArgumentNullException("value");

            byte[] buffer = _encoding.GetBytes(value);

            GetSection().Write(buffer);
            if (nullTerminate)
            {
                GetSection().Write((byte)0);
            }
        }
    }
}
