/**
 * @file    JObjectFormatter.cs
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
using MessagePack;
using MessagePack.Formatters;
using MessagePack.Resolvers;
using Newtonsoft.Json.Linq;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace GoPxLSdk
{
    internal class JObjectFormatter : IMessagePackFormatter<JObject>
    {
        /// <summary>
        /// This function is intended for use by unit tests, and will use a premade msgpack resolver
        /// </summary>
        public static byte[] SerializeToByteArray(JObject jObject)
        {
            var resolver = CompositeResolver.Create(
                // enable extension packages first
                JObjectResolver.Instance,
                // finally use standard (default) resolver
                StandardResolver.Instance
            );
            var options = MessagePackSerializerOptions.Standard.WithResolver(resolver);

            return MessagePackSerializer.Serialize(jObject, options);
        }

        /// <summary>
        /// This function is intended for use by unit tests, and will use a premade msgpack resolver
        /// </summary>
        public static JObject DeserializeFromByteArray(byte[] buffer)
        {
            var resolver = CompositeResolver.Create(
                          // enable extension packages first
                          JObjectResolver.Instance,
                          // finally use standard (default) resolver
                          StandardResolver.Instance
                      );
            var options = MessagePackSerializerOptions.Standard.WithResolver(resolver);

            using (var ms = new MemoryStream(buffer))
            {
                return MessagePackSerializer.Deserialize<JObject>(ms, options);
            }
        }

        public JObject Deserialize(ref MessagePackReader reader, MessagePackSerializerOptions options)
        {
            if (reader.TryReadNil())
            {
                return null;
            }
            else
            {
                // Let the default formatter handle reading into map
                var formatter = options.Resolver.GetFormatter<Dictionary<string, object>>();
                var dict = formatter.Deserialize(ref reader, options);

                // Convert the map into a JObject
                return JObject.FromObject(dict);
            }
        }

        public void Serialize(ref MessagePackWriter writer, JObject value, MessagePackSerializerOptions options)
        {
            if (value == null)
            {
                writer.WriteNil();
            }
            else
            {
                // Collapse the JObject into a Dictionary<string, object>
                var dict = ToDictionary(value);

                // Let the default formatter handle writing out the msgpack map
                var formatter = options.Resolver.GetFormatter<IDictionary<string, object>>();
                formatter.Serialize(ref writer, dict, options);
            }
        }

        /// <summary>
        /// Converts a JObject into a Dictionary
        /// </summary>
        private static IDictionary<string, object> ToDictionary(JObject jObj)
        {
            var propertyValuePairs = jObj.ToObject<Dictionary<string, object>>()
                .OrderBy(kvp => kvp.Key)
                .ToDictionary(kvp => kvp.Key, kvp => kvp.Value);

            ProcessJObjectProperties(propertyValuePairs);
            ProcessJArrayProperties(propertyValuePairs);
            return propertyValuePairs;
        }

        private static void ProcessJObjectProperties(IDictionary<string, object> propertyValuePairs)
        {
            var objectPropertyNames = propertyValuePairs
                .Where(kvp => kvp.Value is JObject)
                .Select(kvp => kvp.Key)
                .ToList();

            objectPropertyNames.ForEach(propertyName => propertyValuePairs[propertyName] = ToDictionary((JObject)propertyValuePairs[propertyName]));
        }

        private static void ProcessJArrayProperties(IDictionary<string, object> propertyValuePairs)
        {
            var arrayPropertyNames = propertyValuePairs
                .Where(kvp => kvp.Value is JArray)
                .Select(kvp => kvp.Key)
                .ToList();

            arrayPropertyNames.ForEach(propertyName => propertyValuePairs[propertyName] = ToArray((JArray)propertyValuePairs[propertyName]));
        }

        private static object ToArray(JArray array)
        {
            return array.ToObject<object[]>().Select(ProcessArrayEntry).ToArray();
        }

        private static object ProcessArrayEntry(object value)
        {
            if (value is JObject jobj)
            {
                return ToDictionary(jobj);
            }
            else if (value is JArray jarr)
            {
                return ToArray(jarr);
            }
            else
            {
                return value;
            }
        }
    }
}
