/**
 * @file    GoCtrlChannelV1.cs
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
using MessagePack;
using MessagePack.Resolvers;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Dynamic;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;

namespace GoPxLSdk.Internal.V1
{
    internal class GoCtrlChannelV1 : IGoCtrlChannel
    {
        public enum ProtocolMode
        {
            Json = 0xB001,
            MessagePack = 0xB000
        }

        public class ThreadException
        {
            /// <summary>
            /// Normally null. Read thread will set to an exception if an unhandled exception on thread occurs.
            /// </summary>
            public Exception ReadException { get; internal set; }

            /// <summary>
            /// Normally null. Write thread will set to an exception if an unhandled exception on thread occurs.
            /// </summary>
            public Exception WriteException { get; internal set; }
        }

        // Private constants
        private const int WRITE_THREAD_LOOP_SLEEP_MILLISEC = 100;   // 100 ms.
        private const int READ_THREAD_LOOP_SLEEP_MILLISEC = 100;     // 100 ms.

        /// <summary>Checks if connected to the server.</summary>
        /// <returns>true if connected</returns>
        public bool IsConnected
        {
            get => _isConnected;
            set => _isConnected = value;
        }
        private volatile bool _isConnected = false;

        private ProtocolMode Mode { get; set; } = ProtocolMode.MessagePack;

        /// <summary>Get the server address</summary>
        /// <returns>The server address</returns>
        public IPAddress Address { get; private set; }

        /// <summary>Get the server port</summary>
        /// <returns>The server port</returns>
        public int Port { get; private set; } = Def.GO_PXL_SDK_DEFAULT_CONTROL_PORT;

        /// <summary>
        /// Get the state of the read/write threads
        /// </summary>
        public ThreadException ThreadState { get; } = new ThreadException();

        private readonly object _lockObj = new object();

        private readonly BlockingCollection<byte[]> _sendQueue = new BlockingCollection<byte[]>();
        private readonly WaitHandle _sendQueueSema = new AutoResetEvent(false);

        private readonly Thread _readThread;
        private readonly Thread _writeThread;

        private TcpClient _client;
        private GoBinaryWriter _writer;
        private GoBinaryReader _reader;

        private GoCtrlChannelResponseHandler _responseHandler;

        private readonly MessagePackSerializerOptions _options = MessagePackSerializerOptions.Standard
            .WithResolver(CompositeResolver.Create(
                // enable extension packages first
                JObjectResolver.Instance,
                // finally use standard (default) resolver
                StandardResolver.Instance
            )
        );

        public GoCtrlChannelV1()
        {
            _writeThread = new Thread(new ThreadStart(WriteThreadFunc))
            {
                IsBackground = true
            };

            _readThread = new Thread(new ThreadStart(ReadThreadFunc))
            {
                IsBackground = true
            };
        }

        /// <summary>
        /// Connects to the server using carrier protocol v1.
        /// </summary>
        /// <param name="address">The IP address you intend to connect to.</param>
        /// <param name="port">The port you intend to connect to.</param>
        /// <param name="millisecondsTimeout">(optional) timeout in milliseconds. Default value is <see cref="Def.DEFAULT_TCP_TIMEOUT_MILLISECONDS"/>.</param>
        /// <exception cref="TimeoutException">Thrown if unable to connect.</exception>
        public void Connect(IPAddress address, int port, int millisecondsTimeout = Def.DEFAULT_TCP_TIMEOUT_MILLISECONDS)
        {
            if (IsConnected)
            {
                return;
            }

            Address = address;
            Port = port;

            _client?.Dispose();
            _client = new TcpClient(AddressFamily.InterNetwork)
            {
                ReceiveTimeout = millisecondsTimeout,
            };

            try
            {
                if (!_client.ConnectAsync(Address, Port).Wait(millisecondsTimeout))
                {
                    throw new TimeoutException($"TCP connection failed to connect within allowed timeout of {millisecondsTimeout}ms");
                }

                IsConnected = true;

                _writer = new GoBinaryWriter(_client.GetStream());
                _reader = new GoBinaryReader(_client.GetStream());
            }
            catch
            {
                _reader?.Dispose();
                _writer?.Dispose();
                throw;
            }

            _writeThread.Start();
            _readThread.Start();
        }

        /// <summary>Disconnects from the server.</summary>
        /// <exception cref="ThreadStateException">Thrown if unable to disconnect.</exception>
        public void Disconnect()
        {
            if (!IsConnected)
            {
                return;
            }

            IsConnected = false;

            _reader?.Dispose();
            _reader = null;

            _writer?.Dispose();
            _writer = null;

            _readThread?.Join();
            _writeThread?.Join();

            _sendQueue.Dispose();

            _client?.Close();
            _client = null;
        }

        /// <summary>
        /// Sets a callback to receive responses from the server, using the implemented carrier protocol.
        /// </summary>
        /// <param name="callback"></param>
        public void SetResponseHandler(GoCtrlChannelResponseHandler callback)
        {
            _responseHandler = callback;
        }

        /// <summary>
        /// Sends a request using carrier protocol v1.
        /// </summary>
        /// <exception cref="GoChannelException">GoChannelError if not connected.</exception>
        public void SendRequest(GoRequest request)
        {
            if (!IsConnected)
            {
                throw new GoChannelException("Not connected.");
            }

            byte[] buffer;
            switch (Mode)
            {
                case ProtocolMode.Json:
                    buffer = request.ToJsonArray();
                    break;
                case ProtocolMode.MessagePack:
                    buffer = request.ToMsgPack();
                    break;
                default:
                    throw new NotImplementedException();
            }

            _sendQueue.Add(buffer);
        }

        private void ProcessRequest(byte[] msgpack)
        {
            if (_writer == null)
            {
                return;
            }

            _writer.BeginWrite<uint>(true);

            _writer.Write((ushort)Mode);
            _writer.Write((uint)msgpack.Length);
            _writer.Write(msgpack, 0, msgpack.Length);

            _writer.EndWrite();
            _writer.Flush();
        }

        private void ProcessResponse()
        {
            if (!IsConnected || _reader == null)
            {
                return;
            }

            int status;
            byte[] buffer;
            bool matchingMessageType;
            ushort messageType;

            _reader.BeginRead<uint>(true);
            try
            {
                // Read the carrier message header of the JSON response message.
                messageType = _reader.ReadUInt16();
                matchingMessageType = messageType == (ushort)Mode;

                // This is the status value in the carrier message header of the JSON response message.
                // It is NOT the status of the JSON response message.
                status = _reader.ReadInt32();

                int bufferSize = (int)_reader.ReadUInt32();
                buffer = _reader.ReadBytes(bufferSize);
            }
            finally
            {
                _reader.EndRead();
            }

            // Only read content if the message type is in the mode we're set to, otherwise 'dump' the data
            if (matchingMessageType)
            {
                GoResponse response = null;
                switch (messageType)
                {
                    case (ushort)ProtocolMode.Json:
                        response = ProcessResponseJsonPayload(buffer);
                        break;
                    case (ushort)ProtocolMode.MessagePack:
                        response = ProcessResponseMessagePackPayload(buffer);
                        break;
                }

                _responseHandler?.Invoke(response);
            }
        }

        private GoResponse ProcessResponseJsonPayload(byte[] buffer)
        {
            string json = Encoding.UTF8.GetString(buffer);
            var jobj = JObject.Parse(json);
            if (!jobj.TryGetValue("type", out var jTokenType))
            {
                JObject tempObj = new JObject
                {
                    ["status"] = (int)GoStatus.ErrorFormat
                };

                return new GoResponse(GoResponseType.Request, tempObj);
            }

            var rawJson = jobj as JObject;

            GoResponseType type = jTokenType.ToObject<GoResponseType>();
            GoResponse response;
            switch (type)
            {
                case GoResponseType.Notification:
                    {
                        response = new GoNotificationResponse(rawJson);
                        break;
                    }
                case GoResponseType.Stream:
                    {
                        response = new GoStreamResponse(rawJson);
                        break;
                    }
                case GoResponseType.Request:
                default:
                    {
                        response = new GoRequestResponse(rawJson);
                        break;
                    }
            }

            return response;
        }

        private GoResponse ProcessResponseMessagePackPayload(byte[] buffer)
        {
            JObject raw = MessagePackSerializer.Deserialize<JObject>(buffer, _options);

            GoResponse response;
            string type = raw.Value<string>("type").ToLower();

            if (type == nameof(GoResponseType.Notification).ToLower())
            {
                response = new GoNotificationResponse(raw);
            }
            else if (type == nameof(GoResponseType.Stream).ToLower())
            {
                response = new GoStreamResponse(raw);
            }
            else if (type == nameof(GoResponseType.Request).ToLower())
            {
                response = new GoRequestResponse(raw);
            }
            else
            {
                response = new GoRequestResponse(raw);
            }

            return response;
        }

        public void Dispose()
        {
            Disconnect();
        }

        private void WriteThreadFunc()
        {
            while (IsConnected)
            {
                bool status = _sendQueue.TryTake(out byte[] msgpack, WRITE_THREAD_LOOP_SLEEP_MILLISEC);
                if (status)
                {
                    try
                    {
                        ProcessRequest(msgpack);
                    }
                    catch (Exception ex)
                    {
                        ThreadState.WriteException = ex;

                        // Fall through and continue running the thread loop.
                    }
                }
                else
                {
                    // Don't exit thread if waiting fails.
                    // GoLogError("Timeout occurred while waiting for data in send queue");
                }
            }
        }

        private void ReadThreadFunc()
        {
            while (IsConnected)
            {
                bool status = _client.Client.Poll(READ_THREAD_LOOP_SLEEP_MILLISEC * 1000, SelectMode.SelectRead);
                if (!status)
                {
                    continue;
                }

                try
                {
                    ProcessResponse();
                }
                catch (Exception ex)
                {
                    ThreadState.ReadException = ex;

                    // Fall through and continue running the thread loop.
                }
            }
        }
    }
}