/*
 * @file            MultiSensorLayout.cs
 * Objective        Connect to the main sensor, add a sensor to sensor group, and change the layout of the group.
 * 
 * GoPxLSdk Sample
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 *
 * Licensed under the MIT License.
 * Redistributions of files must retain the above copyright notice.
 */

using GoPxLSdk;
using Newtonsoft.Json.Linq;
using System;
using System.Linq;
using System.Net;

namespace MultiSensorLayout
{
    internal class MultiSensorLayoutSample
    {
        private const string SENSOR_1_IP = "192.168.1.10";
        private const string SENSOR_2_IP = "192.168.1.11";
        private const string SENSOR_1_SERIAL_NO = "62048";
        private const string SENSOR_2_SERIAL_NO = "62984";

        private const string ENGINE_ID = "g2xxx";
        private const string ENGINE_PATH = "/scan/engines/" + ENGINE_ID;
        private const string SCANNERS_PATH = "/scan/engines/" + ENGINE_ID + "/scanners";
        private const string SENSORS_PATH = "/scan/engines/" + ENGINE_ID + "/scanners/scanner-0/sensors";
        private const string VISIBLE_SENSORS_PATH = "/scan/engines/" + ENGINE_ID + "/visibleSensors";
        private const string SENSOR_LAYOUT_PATH = "/scan/engines/" + ENGINE_ID + "/scanners/scanner-0/layout";

        private const int ERROR_STATUS = -1;
        private const int OK_STATUS = 0;
        private const int REST_COMMAND_TIMEOUT = 3_000; // 3 seconds.
        private const int RECEIVE_DATA_TIMEOUT = 3_000;

        private const int GO_PXL_SDK_DEFAULT_CONTROL_PORT = 3600;

        static string ParseGoRequestException(GoRequestException ex)
        {
            JToken error = ex.Response.Payload["errors"][0];
            return error != null ? $"Error: {error["status"]} {error["description"]}" : "";
        }

        static void PrintSensorLayoutInfo(JObject response)
        {
            JToken sensors = response.SelectToken("grid.sensors");
            Console.WriteLine("\n" + new string('*', 32));
            foreach (JObject sensor in sensors.Cast<JObject>())
            {
                Console.WriteLine("Sensor ID: " + sensor["sensorId"]);
                Console.WriteLine("Display Name: " + sensor["displayName"]);
                Console.WriteLine($"Row: {sensor["row"]}, Column: {sensor["column"]}");
                Console.WriteLine("Orientation: " + sensor["orientation"]);
                Console.WriteLine("Multiplexing Bank: " + sensor["multiplexingBank"] + "\n");
            }
        }

        private static int MultiSensorLayout()
        {// Connect to a sensor system.
            IPAddress ipAddress;
            ipAddress = IPAddress.Parse(SENSOR_1_IP);
            using (var system = new GoSystem(ipAddress, GO_PXL_SDK_DEFAULT_CONTROL_PORT))
            {
                try
                {
                    system.Connect();
                    Console.WriteLine($"\nConnected to {ipAddress}:{GO_PXL_SDK_DEFAULT_CONTROL_PORT}");
                }
                catch (Exception ex)
                {
                    Console.WriteLine(
                        ex.Message
                            + " - Connection failed. Check if sensor is powered on and connected "
                            + "also check if you are using the correct IP address and port number."
                    );
                    return ERROR_STATUS;
                }

                JObject response;
                JObject payload;

                // Check if a sensor group exists.
                response = system.Client().Read(ENGINE_PATH).GetResponse(REST_COMMAND_TIMEOUT).Payload;
                if (string.IsNullOrEmpty(response.SelectToken("_embedded.go:scanner")?.ToString()))
                {
                    Console.WriteLine("\nScanner not present, creating scanner");
                    system.Client().Create(SCANNERS_PATH);
                }
                else
                {
                    Console.WriteLine("\nA Sensor Group exists.");
                }

                // Add a sensor to the sensor group. 
                string[] sensorSerials = { SENSOR_2_SERIAL_NO };
                foreach (string serial in sensorSerials)
                {
                    payload = new JObject { ["serialNumber"] = serial };
                    try
                    {
                        system.Client().Create(SENSORS_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
                        Console.WriteLine("\nAdded " + serial + " to the sensor group.");
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine(ex.Message + " - Failed to add sensor " + serial + ". Check the API path & serial number and make sure firmware is compatible and that sensor is not already added.");
                        return ERROR_STATUS;
                    }
                }

                // Output the sensor group.
                response = system.Client().Read(SENSOR_LAYOUT_PATH).GetResponse(REST_COMMAND_TIMEOUT).Payload;
                PrintSensorLayoutInfo(response);

                // Change the sensor group layout.
                payload = response;
                payload.SelectToken("grid.sensors[2].row").Replace(0);
                payload.SelectToken("grid.sensors[2].column").Replace(1);
                payload.SelectToken("grid.sensors[2].orientation").Replace(0);
                try
                {
                    system.Client().Update(SENSOR_LAYOUT_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
                    Console.WriteLine("Updated sensor layout");
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message + " - Failed to update the layout.");
                    return ERROR_STATUS;
                }

                // Output the current layout of the sensors in the sensor group.
                response = system.Client().Read(SENSOR_LAYOUT_PATH).GetResponse(REST_COMMAND_TIMEOUT).Payload;
                PrintSensorLayoutInfo(response);

                // Remove the senors from group.
                const string SENSOR_PATH = "/scan/engines/" + ENGINE_ID + "/scanners/scanner-0/sensors/sensor-0";
                try
                {
                    system.Client().Delete(SENSOR_PATH).CheckResponse(REST_COMMAND_TIMEOUT);
                    Console.WriteLine("Removed " + SENSOR_PATH + " from sensor group.");
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message + "Failed to remove the sensor. Check the API path or the sensor ID.");
                    return ERROR_STATUS;
                }

                system.Disconnect();
                return OK_STATUS;
            }
        }
        static int Main()
        {
            int status = MultiSensorLayout();

            Console.WriteLine("\nPress any key to exit the program...");
            Console.ReadKey(true);
            return status;
        }
    }
}