/*
 * @file            ReceiveMetrics.cs
 * Objective        Connect to a Gocator sensor system and receive metrics.
 * 
 * GoPxLSdk Sample
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 *
 * Licensed under the MIT License.
 * Redistributions of files must retain the above copyright notice.
 */

using GoPxLSdk;
using Newtonsoft.Json.Linq;
using System;
using System.Net;

namespace ReceiveMetrics
{
    internal class ReceiveMetricsSample
    {
        private const string SENSOR_IP = "192.168.1.10";
        private const int PORT = 3600;
        private const string SYSTEM_METRICS_PATH = "/system/metrics";
        private const int ERROR_STATUS = -1;
        private const int OK_STATUS = 0;

        // Callback function for SetStreamHandler().
        private static void OnMetrics(GoStreamResponse notification)
        {
            var content = notification.Payload;
            Console.WriteLine("\nApplication Uptime: " + content["appUpTime"]);
            Console.WriteLine("CPU Cores Usage Average: " + content["cpuCoresUsedAvg"]);
            Console.WriteLine("Memory Capacity: " + content["memCapacity"]);
            Console.WriteLine("Memory Used: " + content["memUsed"]);
            Console.WriteLine("Type: " + notification.Type);
            Console.WriteLine("Status: " + notification.Status);
            Console.WriteLine("StreamId: " + notification.StreamIdentifier);
            Console.WriteLine("StreamStatus: " + notification.StreamStatus);
            Console.WriteLine("Path: " + notification.Path);
        }

        private static int ReceiveMetrics()
        {
            // Connect to a sensor system.
            IPAddress ipAddress;
            ipAddress = IPAddress.Parse(SENSOR_IP);
            using (var system = new GoSystem(ipAddress, PORT))
            {
                try
                {
                    system.Connect();
                    Console.WriteLine($"\nConnected to {ipAddress}:{PORT}");
                }
                catch (Exception ex)
                {
                    Console.WriteLine(
                        ex.Message
                            + " - Connection failed. Check if sensor is powered on and connected "
                            + "also check if you are using the correct IP address and port number."
                    );
                    return ERROR_STATUS;
                }

                // Run the sensor.
                if (system.RunningState() == GoSystem.State.Ready)
                {
                    system.Start();
                    Console.WriteLine("\nSystem started.");
                }

                // Set the callback function and start receiving data.
                Console.WriteLine("\nRunning callback function to receive metrics...");
                Console.WriteLine("Press ENTER to close the connection and exit.");
                try
                {
                    system.Client().SetStreamHandler(OnMetrics);
                    system.Client().StartStream(SYSTEM_METRICS_PATH);
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message + " - Failed to receive data.");
                    return ERROR_STATUS;
                }

                Console.ReadKey(true);
                system.Client().StopStream(SYSTEM_METRICS_PATH);
                system.Stop();
                system.Disconnect();
            }

            return OK_STATUS;
        }

        static int Main()
        {
            int status = ReceiveMetrics();

            if (status == ERROR_STATUS)
            {
                Console.WriteLine("\nPress any key to exit the program...");
                Console.ReadKey(true);
            }
            return status;
        }
    }
}